(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* SCAP                                                        *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: scap.v,v 1.4 2006/07/15 02:56:41 ssg05 Exp $ *)

Require Export optionT.
Require Import MAPT.
Require Import FMapTB.
Require Import machine.

Require Import extac.

Open Local Scope tm_scope.

Definition stpred := state -> Prop.

Definition acpred := state -> state -> Prop.

Inductive cdspec : Type := 
  | mk_cdspec : stpred -> acpred -> cdspec.

Notation " [ p & g ] " := (mk_cdspec p g) 
  (at level 0) : scap_scope.

Open Local Scope scap_scope.

Module CdSpec <: ONE_TYPE.
  Definition B := cdspec.
End CdSpec.

Module CHSpecDef := FuncMapT CdSpec.
Module CHSpec := MapEx CHSpecDef.
(* psi : chspec *)
Definition chspec := CHSpec.map.

(* Fixpoint scap_wf_is (psi : scap_chspec) (p : stpred) (g : acpred) (is : instrseq)  *)
(*   {struct is} : Prop := *)
(*   match is with *)
(*     | ione i => *)
(*       match i with *)
(*         | jr r =>  *)
(*             (forall s, p s -> optp (g s) (nextS i s)) *)
(* (*             (forall s s', p s  *) *)
(* (*               -> exists p'', exists g'', *) *)
(* (*                 ScapCHSpec.lookup psi (gSR s ra) [p'' & g''] *) *)
(* (*                 -> optp p'' (nextS i s) *) *)
(* (*                 /\ optg g'' (nextS i s) s' -> g s s') *) *)
(*         | j f =>  *)
(*             (forall s s', p s *)
(*               -> exists p'', exists g'', *)
(*                 ScapCHSpec.get psi f = SomeT [p'' & g''] *)
(*                 /\ optp p'' (nextS i s) *)
(*                 /\ (optg g'' (nextS i s) s' -> g s s')) *)
(*         | _ => False *)
(*       end *)
(*     | iseq i is' => *)
(*       match i with *)
(*         | jr ra => False *)
(*         | j f => False *)
(*         | blt rs rt f as i => *)
(*           exists p', exists g', *)
(*             (forall s s', p s -> *)
(*               if blt_nat (gSR s rs) (gSR s rt) *)
(*                 then (exists p'', exists g'', *)
(*                   ScapCHSpec.get psi f = SomeT [p'' & g''] *)
(*                   /\ optp p'' (nextS i s) *)
(*                   /\ (optg g'' (nextS i s) s' -> g s s')) *)
(*                 else (optp p' (nextS i s)  *)
(*                   /\ (optg g' (nextS i s) s' -> g s s'))) *)
(*             /\ scap_wf_is psi p' g' is' *)
(*         | bgtz rs f as i => *)
(*           exists p', exists g', *)
(*             (forall s s', p s -> *)
(*               if blt_nat (O) (gSR s rs) *)
(*                 then (exists p'', exists g'', *)
(*                   ScapCHSpec.get psi f = SomeT [p'' & g''] *)
(*                   /\ optp p'' (nextS i s) *)
(*                   /\ (optg g'' (nextS i s) s' -> g s s')) *)
(*                 else (optp p' (nextS i s)  *)
(*                   /\ (optg g' (nextS i s) s' -> g s s'))) *)
(*             /\ scap_wf_is psi p' g' is' *)
(*         | _ as i =>  *)
(*           exists p', exists g', *)
(*             (forall s s', p s  *)
(*               -> optp p' (nextS i s)  *)
(*               /\ (optg g' (nextS i s) s' -> g s s')) *)
(*             /\ scap_wf_is psi p' g' is' *)
(*       end *)
(*   end. *)

Inductive wfis : chspec -> stpred -> acpred -> instrseq -> Prop :=

  | wfis_j :
      forall (psi : chspec) (p : stpred) (g : acpred) f,
        (forall s, p s
          -> exists p'', exists g'' : acpred,
            CHSpec.get psi f = SomeT [p'' & g'']
            /\ (optp p'' (nextS (j f) s))
            /\ (forall s', 
              optg g'' (nextS (j f) s) s' -> g s s'))
        -> wfis psi p g (ione (j f))

  | wfis_call :
      forall (psi : chspec) (p : stpred) (g : acpred) f,
        (forall s, p s 
          -> exists p'', exists g'',
            CHSpec.get psi f = SomeT [p'' & g'']
            /\ optp p'' (nextS (j f) s)
            /\ (forall s', 
              optg g'' (nextS (j f) s) s'
              -> exists p''', exists g''',
                CHSpec.get psi (S_pc s') = SomeT [p''' & g''']
                /\ p''' s'
                /\ (forall s'', g''' s' s'' -> g s s'')))
        -> wfis psi p g (ione (j f))

  | wfis_jr : 
      forall (psi : chspec) (p : stpred) (g : acpred) r,
        (forall s, p s
          -> exists p'', exists g'',
            CHSpec.get psi (rget s r) = SomeT [p'' & g'']
            /\ optp p'' (nextS (jr r) s)
            /\ (forall s', 
              optg g'' (nextS (jr r) s) s' -> g s s'))
        -> wfis psi p g (ione (jr r))

  | wfis_ret :
      forall (psi : chspec) (p : stpred) (g : acpred) r,
        (forall s, p s -> optp (g s) (nextS (jr r) s))
        -> wfis psi p g (ione (jr r))

  | wfis_addu :
      forall (psi : chspec) (p p' : stpred) (g g' : acpred) 
        rd rs rt is',
        (forall s, p s 
          -> optp p' (nextS (addu rd rs rt) s)
          /\ (forall s', 
            optg g' (nextS (addu rd rs rt) s) s' -> g s s'))
        -> wfis psi p' g' is'
        -> wfis psi p g (iseq (addu rd rs rt) is')

  | wfis_addiu :
      forall (psi : chspec) (p p' : stpred) (g g' : acpred) 
        rd rs w is',
        (forall s, p s 
          -> optp p' (nextS (addiu rd rs w) s)
          /\ (forall s', 
            optg g' (nextS (addiu rd rs w) s) s' -> g s s'))
        -> wfis psi p' g' is'
        -> wfis psi p g (iseq (addiu rd rs w) is')

  | wfis_subu :
      forall (psi : chspec) (p p' : stpred) (g g' : acpred) 
        rd rs rt is',
        (forall s, p s 
          -> optp p' (nextS (subu rd rs rt) s)
          /\ (forall s', 
            optg g' (nextS (subu rd rs rt) s) s' -> g s s'))
        -> wfis psi p' g' is'
        -> wfis psi p g (iseq (subu rd rs rt) is')

  | wfis_subiu :
      forall (psi : chspec) (p p' : stpred) (g g' : acpred) 
        rd rs w is',
        (forall s, p s 
          -> optp p' (nextS (subiu rd rs w) s)
          /\ (forall s', 
            optg g' (nextS (subiu rd rs w) s) s' -> g s s'))
        -> wfis psi p' g' is'
        -> wfis psi p g (iseq (subiu rd rs w) is')

  | wfis_lw :
      forall (psi : chspec) (p p' : stpred) (g g' : acpred) 
        rt w rs is',
        (forall s, p s 
          -> optp p' (nextS (lw rt w rs) s)
          /\ (forall s', 
            optg g' (nextS (lw rt w rs) s) s' -> g s s'))
        -> wfis psi p' g' is'
        -> wfis psi p g (iseq (lw rt w rs) is')

  | wfis_sw :
      forall (psi : chspec) (p p' : stpred) (g g' : acpred) 
        rt w rs is',
        (forall s, p s 
          -> optp p' (nextS (sw rt w rs) s)
          /\ (forall s', 
            optg g' (nextS (sw rt w rs) s) s' -> g s s'))
        -> wfis psi p' g' is'
        -> wfis psi p g (iseq (sw rt w rs) is')
  | wfis_blt :
      forall (psi : chspec) (p p' : stpred) (g g' : acpred) 
        rs rt f is',
        (forall s, p s -> if blt_nat (rget s rs) (rget s rt) 
          then 
            (exists p'', exists g'',
              CHSpec.get psi f = SomeT [p'' & g'']
              /\ optp p'' (nextS (blt rs rt f) s)
              /\ forall s', 
                optg g'' (nextS (blt rs rt f) s) s' -> g s s')
          else
            optp p' (nextS (blt rs rt f) s)
            /\ forall s',
              optg g' (nextS (blt rs rt f) s) s' -> g s s')
        -> wfis psi p' g' is'
        -> wfis psi p g (iseq (blt rs rt f) is')

  | wfis_bgtz :
      forall (psi : chspec) (p p' : stpred) (g g' : acpred) 
        rs f is',
        (forall s, p s -> if blt_nat O (rget s rs) 
          then 
            (exists p'', exists g'',
              CHSpec.get psi f = SomeT [p'' & g'']
              /\ optp p'' (nextS (bgtz rs f) s)
              /\ forall s', 
                optg g'' (nextS (bgtz rs f) s) s' -> g s s')
          else
            optp p' (nextS (bgtz rs f) s)
            /\ forall s',
              optg g' (nextS (bgtz rs f) s) s' -> g s s')
        -> wfis psi p' g' is'
        -> wfis psi p g (iseq (bgtz rs f) is')
.

Lemma wfis_weakening : 
  forall is psi (p p' : stpred) (g g' : acpred),
    (forall s, p s -> p' s /\ forall s', g' s s' -> g s s') 
    -> wfis psi p' g' is 
    -> wfis psi p g is.
Proof.
  induction is.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    intros psi p p0 g g0 Hw His;
      try inversion_clear His.

  apply wfis_j.
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH H with (H s Hw1).
  destruct H as [p'' [g'' [Hin'' [Hp'' Hg'']]]].
  exists p''; exists g''.
  split_l.
  split_l.
  intros s' Hg''s'.
  substH Hg'' with (Hg'' s' Hg''s').
  apply (Hw2 s' Hg'').

  apply wfis_call.
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH H with (H s Hw1).
  destruct H as [p'' [g'' [Hin'' [Hp'' Hg'']]]].
  exists p''; exists g''.
  split_l.
  split_l.
  intros s' Hg''s'.
  substH Hg'' with (Hg'' s' Hg''s').
  destruct Hg'' as [p''' [g''' 
    [Hin''' [Hp''' Hg''']]]].
  exists p'''; exists g'''.
  split_l.
  split_l.
  intros s'' H.
  apply Hw2.
  apply Hg'''; trivial.

  apply wfis_jr.
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH H with (H s Hw1).
  destruct H as [p'' [g'' [Hin'' [Hp'' Hg'']]]].
  exists p''; exists g''.
  split_l.
  split_l.
  intros s' Hg''s'.
  substH Hg'' with (Hg'' s' Hg''s').
  apply (Hw2 s' Hg'').

  apply wfis_ret.
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH H with (H s Hw1).
  destruct (nextS (jr r) s); [idtac|destruct H].
  apply Hw2; trivial.

  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    intros psi p p0 g g0 Hw His;
      try inversion_clear His.

  apply wfis_addu with p' g'.
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH H with (H s Hw1).
  destruct H as [Hp' Hg'].
  split_l.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  apply H0.

  apply wfis_addiu with p' g'.
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH H with (H s Hw1).
  destruct H as [Hp' Hg'].
  split_l.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  apply H0.

  apply wfis_subu with p' g'.
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH H with (H s Hw1).
  destruct H as [Hp' Hg'].
  split_l.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  apply H0.

  apply wfis_subiu with p' g'.
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH H with (H s Hw1).
  destruct H as [Hp' Hg'].
  split_l.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  apply H0.

  apply wfis_lw with p' g'.
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH H with (H s Hw1).
  destruct H as [Hp' Hg'].
  split_l.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  apply H0.

  apply wfis_sw with p' g'.
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH H with (H s Hw1).
  destruct H as [Hp' Hg'].
  split_l.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  apply H0.

  apply wfis_blt with p' g'.
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH H with (H s Hw1).
  bnat_blt_dec (rget s rs) (rget s rt).
  (* rs < rt *)
  destruct H as [p'' [g'' [Hin'' [Hp'' Hg'']]]].
  exists p''; exists g''.
  split_l.
  split_l.
  intros s' Hg''s'.
  substH Hg'' with (Hg'' s' Hg''s').
  apply (Hw2 s' Hg'').
  (* rs >= rt*)
  destruct H as [Hp' Hg'].
  split_l.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  apply H0.

  apply wfis_bgtz with p' g'.
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH H with (H s Hw1).
  bnat_blt_dec O (rget s rs).
  (* rs < rt *)
  destruct H as [p'' [g'' [Hin'' [Hp'' Hg'']]]].
  exists p''; exists g''.
  split_l.
  split_l.
  intros s' Hg''s'.
  substH Hg'' with (Hg'' s' Hg''s').
  apply (Hw2 s' Hg'').
  (* rs >= rt*)
  destruct H as [Hp' Hg'].
  split_l.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  apply H0.
Qed.

Definition wfch (psi psi' : chspec) (ch : codeheap) :=
  forall f p g,
    CHSpec.lookup psi' f (mk_cdspec p g)
    -> (exists is : _, codeheap.getis ch f is
      /\ wfis psi p g is).

(* 
   This vcgen function treat jr as ret and j as jump (not call!!)

  TODO: add one more param to tell vcgen how to treat jr and j.

*)

Fixpoint vc (psi : chspec) (is : instrseq) 
  {struct is} : optionT cdspec :=
  match is with
    | ione (jr r) => 
          SomeT (mk_cdspec 
            (fun s => True)
            (fun s => fun s' => 
              optp (fun s'' => s'' = s') (nextS2 (jr r) s)))
    | ione (j f) => 
      match (CHSpec.get psi f) with
        | SomeT (mk_cdspec p' g') =>
          SomeT (mk_cdspec
            (fun s => p' (pcset s f))
            (fun s => fun s'' => g' (pcset s f) s''))
        | _ => NoneT
      end
    | iseq (blt rs rt f) is' =>
      match vc psi is' with
        | SomeT (mk_cdspec p' g') =>
          SomeT (mk_cdspec
            (fun s => 
              if blt_nat (rget s rs) (rget s rt)
                then 
                  (optpT (fun bsp : cdspec => let (p'', _) := bsp in 
                    (p'' (pcset s f)))
                    (CHSpec.get psi f))
                else (p' (pcinc s)))
            (fun s => fun s' =>
              if blt_nat (rget s rs) (rget s rt) 
                then
                  (opt_predT (CHSpec.get psi f) (fun bsp =>
                    let (_, g'') := bsp in 
                      (g'' (pcset s f) s')))
                else (g' (pcinc s) s')))
        | NoneT => NoneT
      end
    | iseq (bgtz rs f) is' =>
      match vc psi is' with
        | SomeT (mk_cdspec p' g') =>
          SomeT (mk_cdspec
            (fun s => 
              if blt_nat O (rget s rs)
                then 
                  (optpT (fun bsp : cdspec => let (p'', _) := bsp in 
                    (p'' (pcset s f)))
                    (CHSpec.get psi f))
                else (p' (pcinc s)))
            (fun s => fun s' =>
              if blt_nat O (rget s rs) 
                then
                  (opt_predT (CHSpec.get psi f) (fun bsp =>
                    let (_, g'') := bsp in 
                      (g'' (pcset s f) s')))
                else (g' (pcinc s) s')))
        | NoneT => NoneT
      end
    | iseq (j f) is' => NoneT
    | iseq (jr r) is' => NoneT
    | iseq i is' =>
      match vc psi is' with
        | SomeT (mk_cdspec p' g') =>
          SomeT (mk_cdspec
            (fun s => 
              optp (fun s' => p' s') (nextS2 i s))
            (fun s => fun s'' => 
              optp (fun s' => g' s' s'') (nextS2 i s)))
        | NoneT => NoneT
      end
    | _ => NoneT
  end.

(* Fixpoint ra_eq_pred (is : iseq) (P : chspec) {struct is} : Prop := *)
(*   match is with *)
(*     | iseq i is' => ra_eq_pred is' P *)
(* (*     | jal fret target =>  *) *)
(* (*       match P target with *) *)
(* (*         | SomeT (bspeci p g) =>  *) *)
(* (*           forall s s', g s s' -> rget s ra = rget s' ra *) *)
(* (*         | NoneT => True *) *)
(* (*       end *) *)
(*     | _ => True *)
(*   end. *)

Theorem vc_sound : forall psi is p g,
  (vc psi is) = SomeT [ p & g ] ->
  wfis psi p g is.
Proof.
  intros psi.
  induction is.
  intros p g.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    simpl; intro H; try discriminate.

  (* case i = j f *)
  apply wfis_j.
  intros s Hp.
  destruct (CHSpec.get psi f) as [[p'' g''] | ]; 
    [idtac | discriminate H].
  exists p''; exists g''.
  split_l.
  injection H; intros; subst p; subst g; clear H.
  destruct s as [h rf pc].
  split_lr.

  (* case i = jr r *)
  apply wfis_ret.
  intros s Hps.
  destruct s as [h rf pc].
  injection H; intros; subst p; subst g; clear H.
  simpl.
  trivial.

  (* case is = iseq i is' *)
  intros p g.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    intro H; try discriminate.
  
  (* case i = addu rd rs rt *)
  simpl in H.
  destruct (vc psi is) as [[p' g'] | ]; [idtac|discriminate H]. 
  injection H; intros; subst p; subst g; clear H.
  apply wfis_addu with p' g'.
  simpl.
  destruct s as [h rf pc].
  unfoldgs.
  split_lr.
  apply IHis; trivial.

  (* case i = addiu rd rs w *)
  simpl in H.
  destruct (vc psi is) as [[p' g'] | ]; [idtac|discriminate H]. 
  injection H; intros; subst p; subst g; clear H.
  apply wfis_addiu with p' g'.
  simpl.
  destruct s as [h rf pc].
  split_lr.
  apply IHis; trivial.

  (* case i = subu rd rs rt *)
  simpl in H.
  destruct (vc psi is) as [[p' g'] | ]; [idtac|discriminate H]. 
  injection H; intros; subst p; subst g; clear H.
  apply wfis_subu with p' g'.
  simpl.
  destruct s as [h rf pc].
  split_lr.
  apply IHis; trivial.

  (* case i = subiu rd rs w *)
  simpl in H.
  destruct (vc psi is) as [[p' g'] | ]; [idtac|discriminate H]. 
  injection H; intros; subst p; subst g; clear H.
  apply wfis_subiu with p' g'.
  simpl.
  destruct s as [h rf pc].
  split_lr.
  apply IHis; trivial.

  (* case i = lw rt w rs *)
  simpl in H.
  destruct (vc psi is) as [[p' g'] | ]; [idtac|discriminate H]. 
  injection H; intros; subst p; subst g; clear H.
  apply wfis_lw with p' g'.
  simpl.
  destruct s as [h rf pc].
  split_lr.
  apply IHis; trivial.

  (* case i = sw rt w rs *)
  simpl in H.
  destruct (vc psi is) as [[p' g'] | ]; [idtac|discriminate H]. 
  injection H; intros; subst p; subst g; clear H.
  apply wfis_sw with p' g'.
  simpl.
  destruct s as [h rf pc].
  split_lr.
  apply IHis; trivial.

  (* case i = blt rs rt f *)
  simpl in H.
  destruct (vc psi is) as [[p' g'] | ]; [idtac|discriminate].
  injection H; intros; subst p; subst g; clear H.
  apply wfis_blt with p' g'.
  intros s.
  destruct s as [h rf pc].
  simpl.
  bnat_blt_dec (gR rf rs) (gR rf rt).
  intro H.
  destruct (CHSpec.get psi f) as [[p'' g''] | ]; 
    [idtac | destruct H].
  exists p''; exists g''.
  split_l.
  split_lr.
  intro H.
  split_lr.
  apply IHis; trivial.

  (* case i = bgtz rs rt f *)
  simpl in H.
  destruct (vc psi is) as [[p' g'] | ]; [idtac|discriminate].
  unfold rget in * .
  injection H; intros; subst p; subst g; clear H.
  apply wfis_bgtz with p' g'.
  intros s.
  destruct s as [h rf pc].
  unfold nextS.
  unfold rget.
  bnat_blt_dec O (gR rf rs).
  simpl in Hblt.
  rewrite Hblt.
  intro H.
  destruct (CHSpec.get psi f) as [[p'' g''] | ]; 
    [idtac | destruct H].
  exists p''; exists g''.
  split_l.
  split_lr.
  simpl in Hnblt.
  rewrite Hnblt.

  intro H.
  split_lr.
  apply IHis; trivial.
Qed.

Ltac simplpg :=
  match goal with
    | [ |- exists p, exists g, 
      SomeT (mk_cdspec ?p0 ?g0) = SomeT (mk_cdspec p g)] =>
    exists p0; exists g0; reflexivity
  end.

Ltac simplvc :=
  match goal with
    | [ |- wfis ?Psi ?pre ?guar ?is ] =>
      let E := fresh "E" in let p := fresh "p" in let g := fresh "g" in
        let E1 := fresh "E" in let E2 := fresh "E" in
          (assert (E : exists p, exists g, 
            vc Psi is = SomeT (mk_cdspec p g));
          [simpl; simplpg | destruct E as [p [g E]]];
          apply wfis_weakening with (p' := p) (g' := g);
            [ injection E; intros E1 E2; subst p; subst g; clear E 
              |
              apply vc_sound; simpl; auto])
  end.
