(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* SCAP over OCAP                                               *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: scap_ocap.v,v 1.4 2006/07/15 02:56:41 ssg05 Exp $ *)

Require Import machine.
Require Import scap.
Require Import ocap.

Require Import extac.
Require Import deptac.

Open Local Scope tm_scope.
Open Local Scope ocap_scope.
Open Local Scope scap_scope.

Module Type SCAPPARAM.
Parameter rho_scap : nat.
  
Parameter D0 : idict.

Axiom rho_scap_not_indom_D0 :
  ~ DSpec.indom D0 rho_scap.
End SCAPPARAM.

Module ScapOcap (Params : SCAPPARAM).

Import Params.

Notation Lscap := scap.cdspec.
Notation mk_Lscap := scap.mk_cdspec.
Notation scap_chspec := scap.chspec.
Notation scap_wfis := scap.wfis.
Notation ocap_wfis := ocap.wfis.
Notation scap_wfch := scap.wfch.
Notation ocap_wfch := ocap.WF_C.

Module ScapCHSpec := scap.CHSpec.

(* *********************************************************** *)

Definition scap_chspec_lift (psi : scap_chspec) (OPsi : chspec) : Prop :=
  (forall f p g,
    ScapCHSpec.lookup psi f [p & g] 
    -> OCHSpec.lookup OPsi f <* rho_scap, [p & g] *>)
  /\ (forall f pi, 
    OCHSpec.lookup OPsi f pi
    -> exists p, exists g, 
      pi = <* rho_scap, [p & g] *>
      /\ ScapCHSpec.lookup psi f [p & g]).

(* *********************************************************** *)
Fixpoint scapWFST (n : nat) (g : acpred) (s : state)
  (psi : chspec) {struct n} : Prop :=
  match n with
    | O => forall s', g s s' ->
      exists pi : ocdspec,
        OCHSpec.lookup psi (S_pc s') pi
        /\ {[ pi | D0 ]} psi s'
    | S n => forall s' : state, g s s' ->
      exists p' : stpred, exists g' : acpred,
        OCHSpec.lookup psi (S_pc s')
            (mk_pi rho_scap (xcdspec_pack [p' & g']))
       /\ p' s'
       /\ scapWFST n g' s' psi
  end.

Definition itpscap (theta : Lscap) : assert := 
  let (p, g) := theta in
    fun (psi : chspec) (s : state)  =>
      p s /\ 
      exists n, scapWFST n g s psi.

Notation "[| theta |]" := (itpscap theta) 
  (at level 0) : ocap_scope.

Notation "[| p & g |]" := (itpscap (mk_Lscap p g))
  (at level 0) : ocap_scope.

Definition D_scap : idict := fun id =>
  if (beq_nat id rho_scap) then SomeT <@ itpscap @> else
    NoneT.

Lemma D_scap_disj_D0 :
  DSpec.disj D_scap D0.
Proof.
  apply DSpec.indom_disj.
  intro a.
  split; intros H1 H2.
  unfold DSpec.indom, D_scap in H1.
  destruct H1 as [b H1].
  unfold DSpec.get, DSpecDef.get in H1.
  beq_nat_dec a rho_scap.
  rewrite (beq_eq a rho_scap Hbeq) in H2.
  destruct (rho_scap_not_indom_D0 H2).
  rewrite Hbneq in H1.
  discriminate H1.
  unfold DSpec.indom, D_scap in H2.
  destruct H2 as [b H2].
  unfold DSpec.get, DSpecDef.get in H2.
  beq_nat_dec a rho_scap.
  rewrite (beq_eq a rho_scap Hbeq) in H1.
  destruct (rho_scap_not_indom_D0 H1).
  rewrite Hbneq in H2.
  discriminate H2.
Qed.

Definition D1 := DSpec.merge D0 D_scap.

Lemma D0_sub_D1 : DSpec.sub D0 D1.
Proof.
  unfold D1.
  apply DSpec.sub_merge_sub_l.
  apply DSpec.disj_compat.
  apply DSpec.disj_sym.
  apply D_scap_disj_D0.
Qed.

Lemma D_scap_sub_D1 : DSpec.sub D_scap D1.
Proof.
  unfold D1.
  apply DSpec.sub_merge_sub_r.
  apply DSpec.disj_compat.
  apply DSpec.disj_sym.
  apply D_scap_disj_D0.
Qed.

Lemma stack_strengthen :
  forall psi n (g g' : acpred) s s',
    scapWFST n g s psi
    -> (forall s'', g' s' s'' -> g s s'')
    -> scapWFST n g' s' psi.
Proof.
  intros.
  induction n.
  unfold scapWFST in *.
  intros s'' Hg'.
  substH H0 with (H0 s'' Hg').
  substH H with (H s'' H0).
  apply H.
  simpl in *.
  intros s'' Hg'.
  substH H0 with (H0 s'' Hg').
  substH H with (H s'' H0).
  apply H.
Qed.

Lemma scap_call_3 : forall (D : idict) (psi : chspec)
  (p p' p'': stpred) (g g' g'': acpred) (f : label),
  DSpec.lookup D rho_scap <@itpscap@>
  -> OCHSpec.lookup psi f <* rho_scap, [p' & g'] *>
  -> (forall s, p s -> optp p' (nextS (j f) s))
  -> (forall s s', p s
    -> optg g' (nextS (j f) s) s'
    -> (OCHSpec.lookup psi (S_pc s')
        <* rho_scap, [p'' & g''] *>
        /\ p'' s'
        /\ (forall s'', g'' s' s'' -> g s s'')))
  -> wfis D <<[|p & g|] / psi>> (ione (j f)).
Proof.
  intros D psi p p' p'' g g' g'' f Hd Hin Hp' Hg'.
  apply wfis_j. unfold lift, itpscap.
  intros psi0 s [Hsub [Hp Hk]].
  simpl in *.
  unfold BigInterp.
  exists <* rho_scap, [p' & g'] *>.
  simpl in *.
  substH Hp' with (Hp' s Hp).
  substH Hg' with (Hg' s).
  assert (H2 : OCHSpec.lookup psi0 f <*rho_scap, [p' & g']*>).
  apply (OCHSpec.lookup_sub_lookup Hin Hsub).
  split_l.
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  split_l.
  destruct Hk as [n Hk].
  exists (S n).
  simpl.
  intros s' Hg'0.
  substH Hg' with (Hg' s' Hp Hg'0).
  exists p''. exists g''.
  destruct Hg' as [Hin' [Hp'' Hg'']].
  split_l using 
    apply (OCHSpec.lookup_sub_lookup Hin' Hsub).
  split_l.
  apply (stack_strengthen psi0 n g g'' s s'
    Hk Hg'').
Qed.

Lemma scap_call : forall (D : idict) (OPsi : chspec)
  (p : stpred) (g : acpred) (f : label),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> exists p'', exists g'', 
      OCHSpec.lookup OPsi f <* rho_scap, [p'' & g''] *>
      /\ optp p'' (nextS (j f) s)
      /\ (forall s', 
        optg g'' (nextS (j f) s) s'
        -> exists p''', exists g''',
          OCHSpec.lookup OPsi (S_pc s') <* rho_scap, [p''' & g''']*>
          /\ p''' s'
          /\ (forall s'', g''' s' s'' -> g s s'')))
  -> wfis D <<[|p & g|] / OPsi>> (ione (j f)).
Proof.
  intros D OPsi p g f Hd Himp.
  apply wfis_j. unfold lift, itpscap.
  intros Psi s [Hsub [Hp Hk]].
  simpl in *.
  unfold BigInterp.
  substH Himp with (Himp s Hp).
  destruct Himp as [p'' [g'' Himp]].
  exists <* rho_scap, [p'' & g''] *>.
  destruct Himp as [Hin Himp].
  split_l using 
    (apply (OCHSpec.lookup_sub_lookup Hin Hsub)).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  destruct Himp as [Hp'' Himp].
  split_l.

  destruct Hk as [n Hk].
  exists (S n).
  simpl.
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  destruct Himp as [p''' [g''' Himp]].
  exists p'''; exists g'''.
  destruct Himp as [Hin' [Hp''' Hg''']].
  split_l using
    (apply (OCHSpec.lookup_sub_lookup Hin' Hsub)).
  split_l.
  apply (stack_strengthen Psi n g g''' s s'
    Hk Hg''').
Qed.

Lemma scap_call_2 : forall (D : idict) (OPsi : chspec)
  (p : stpred) (g : acpred) (f : label),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> exists p'', exists g'', 
      OCHSpec.lookup OPsi f <* rho_scap, [p'' & g''] *>
      /\ optp p'' (nextS (j f) s)
      /\ (forall s', 
        optg g'' (nextS (j f) s) s'
        -> exists p''', exists g''',
          OCHSpec.lookup OPsi (S_pc s') <* rho_scap, [p''' & g''']*>
          /\ p''' s'
          /\ (forall s'', g''' s' s'' -> g s s'')))
  -> wfis D << {[ <* rho_scap, [p & g] *> | D ]} / OPsi>> (ione (j f)).
Proof.
  intros D OPsi p g f Hd Himp.
  simpl.
  apply wfis_j. unfold lift, itpscap.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  simpl in *.
  unfold BigInterp.
  substH Himp with (Himp s Hp).
  destruct Himp as [p'' [g'' Himp]].
  exists <* rho_scap, [p'' & g''] *>.
  destruct Himp as [Hin Himp].
  split_l using 
    (apply (OCHSpec.lookup_sub_lookup Hin Hsub)).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  destruct Himp as [Hp'' Himp].
  split_l.

  destruct Hk as [n Hk].
  exists (S n).
  simpl.
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  destruct Himp as [p''' [g''' Himp]].
  exists p'''; exists g'''.
  destruct Himp as [Hin' [Hp''' Hg''']].
  split_l using
    (apply (OCHSpec.lookup_sub_lookup Hin' Hsub)).
  split_l.
  apply (stack_strengthen Psi n g g''' s s'
    Hk Hg''').
Qed.

Lemma scap_tcall_3 : forall (D : idict) (psi : chspec)
  (p p' : stpred) (g g' : acpred) (f : label),
  DSpec.lookup D rho_scap <@itpscap@>
  -> OCHSpec.lookup psi f <* rho_scap, [p' & g'] *>
  -> (forall s, p s -> optp p' (nextS (j f) s))
  -> (forall s s', p s -> optg g' (nextS (j f) s) s' -> g s s')
  -> wfis D <<[|p & g|] / psi>> (ione (j f)).
Proof.
  intros D psi p p' g g' f Hd Hin Hp' Hg'.
  apply wfis_j. unfold lift, itpscap.
  intros psi0 s [Hsub [Hp Hk]].
  exists (<* rho_scap, [p' & g'] *>).
  split_l using (apply (OCHSpec.lookup_sub_lookup Hin Hsub)).
  substH Hp' with (Hp' s Hp).
  substH Hg' with (Hg' s).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl in * .
  exists itpscap.
  split_l.
  unfold itpscap.
  split_l.
  destruct Hk as [n Hk].
  exists n.
  destruct n.
  simpl in * .
  intros s' Hg's0.
  substH Hg' with (Hg' s' Hp Hg's0).
  substH Hk with (Hk s' Hg').
  trivial.
  simpl in * .
  intros s' Hg's0.
  substH Hg' with (Hg' s' Hp Hg's0).
  substH Hk with (Hk s' Hg').
  trivial.
Qed.

Lemma scap_tcall : forall (D : idict) (OPsi : chspec)
  (p : stpred) (g : acpred) (f : label),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s
    -> exists p'', exists g'',
      OCHSpec.lookup OPsi f <* rho_scap, [p'' & g''] *>
      /\ optp p'' (nextS (j f) s)
      /\ (forall s', optg g'' (nextS (j f) s) s' -> g s s'))
  -> wfis D <<[|p & g|] / OPsi>> (ione (j f)).
Proof.
  intros D OPsi p g f Hd Himp.
  apply wfis_j.
  unfold lift, itpscap, BigInterp.
  intros Psi s [Hsub [Hp Hk]].
  substH Himp with (Himp s Hp).
  destruct Himp as [p'' [g'' Himp]].
  exists <* rho_scap, [p'' & g''] *>.
  destruct Himp as [Hin Himp].
  split_l using 
    (apply (OCHSpec.lookup_sub_lookup Hin Hsub)).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  destruct Himp as [Hp'' Himp].
  split_l.

  destruct Hk as [n Hk].
  exists n.
  destruct n.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
Qed.

Lemma scap_tcall_2 : forall (D : idict) (OPsi : chspec)
  (p : stpred) (g : acpred) (f : label),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s
    -> exists p'', exists g'',
      OCHSpec.lookup OPsi f <* rho_scap, [p'' & g''] *>
      /\ optp p'' (nextS (j f) s)
      /\ (forall s', optg g'' (nextS (j f) s) s' -> g s s'))
  -> wfis D << {[ <* rho_scap, [p & g] *> | D]} / OPsi>> 
  (ione (j f)).
Proof.
  intros D OPsi p g f Hd Himp.
  apply wfis_j.
  unfold lift, itpscap, BigInterp.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Himp with (Himp s Hp).
  destruct Himp as [p'' [g'' Himp]].
  exists <* rho_scap, [p'' & g''] *>.
  destruct Himp as [Hin Himp].
  split_l using 
    (apply (OCHSpec.lookup_sub_lookup Hin Hsub)).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  destruct Himp as [Hp'' Himp].
  split_l.

  destruct Hk as [n Hk].
  exists n.
  destruct n.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
Qed.

Lemma scap_ret : forall (D : idict) (psi : chspec)
  (p : stpred) (g : acpred) (ra : reg),
  DSpec.sub D0 D
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s -> optp (g s) (nextS (jr ra) s))
  -> wfis D <<[| p & g|] /psi>> (ione (jr ra)).
Proof.
  intros D psi p g ra HDsub Hd Hs.
  apply wfis_jr.  unfold lift, itpscap.
  intros psi0 s [Hsub [Hp Hk]].
  substH Hs with (Hs s Hp) into Hg.
  destruct Hk as [n Hk].
  destruct n.
  (* case n = 0 *)
  simpl in * .
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  set (s' := <.h, rf, gR rf ra.>).
  substH Hk with (Hk s' Hg).
  destruct Hk as [pi [Hfin H]].
  simpl in * .
  exists pi.
  split_l.
  apply (biginterp_enlarge_D HDsub psi0 s' H).
  (* case n > 0 *)
  simpl in * .
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  set (s' := <.h, rf, gR rf ra.>).
  substH Hk with (Hk s' Hg).
  destruct Hk as [p' [g' Hk]].
  destruct Hk as [Hin [Hp's' Hk]].
  exists <* rho_scap, [p' & g'] *> .
  simpl in * .
  split_l.
  exists itpscap.
  split_l.
  unfold itpscap.
  split_l.
  exists n.
  assumption.
Qed.

Lemma scap_ret_2 : forall (D : idict) (OPsi : chspec)
  (p : stpred) (g : acpred) (ra : reg),
  DSpec.sub D0 D
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s -> optp (g s) (nextS (jr ra) s))
  -> wfis D << {[ <* rho_scap, [ p & g] *> | D]} /OPsi>> 
  (ione (jr ra)).
Proof.
  intros D OPsi p g ra HDsub Hd Hs.
  apply wfis_jr.  unfold lift, itpscap.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Hs with (Hs s Hp) into Hg.
  destruct Hk as [n Hk].
  destruct n.
  (* case n = 0 *)
  simpl in * .
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  set (s' := <.h, rf, gR rf ra.>).
  substH Hk with (Hk s' Hg).
  destruct Hk as [pi [Hfin H]].
  simpl in * .
  exists pi.
  split_l.
  apply (biginterp_enlarge_D HDsub Psi s' H).
  (* case n > 0 *)
  simpl in * .
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  set (s' := <.h, rf, gR rf ra.>).
  substH Hk with (Hk s' Hg).
  destruct Hk as [p' [g' Hk]].
  destruct Hk as [Hin [Hp's' Hk]].
  exists <* rho_scap, [p' & g'] *> .
  simpl in * .
  split_l.
  exists itpscap.
  split_l.
  unfold itpscap.
  split_l.
  exists n.
  assumption.
Qed.

Lemma scap_jr : forall (D : idict) (OPsi : chspec)
  (p : stpred) (g : acpred) (r : reg),
  DSpec.sub D0 D
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s
    -> exists p'', exists g'',
      OCHSpec.lookup OPsi (gSR s r) <* rho_scap, [p'' & g''] *>
      /\ optp p'' (nextS (jr r) s)
      /\ (forall s', 
        optg g'' (nextS (jr r) s) s' -> g s s'))
  -> wfis D <<[| p & g|] /OPsi>> (ione (jr r)).
Proof.
  intros D OPsi p g r HDsub Hd Himp.
  apply wfis_jr.  
  unfold lift, itpscap, BigInterp.
  intros Psi s [Hsub [Hp Hk]].
  substH Himp with (Himp s Hp).
  destruct Himp as [p'' [g'' Himp]].
  exists <* rho_scap, [p'' & g''] *>.
  destruct Himp as [Hin Himp].
  split_l using 
    (apply (OCHSpec.lookup_sub_lookup Hin Hsub)).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  destruct Himp as [Hp'' Himp].
  split_l.

  destruct Hk as [n Hk].
  exists n.
  destruct n.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
Qed.

Lemma scap_jr_2 : forall (D : idict) (OPsi : chspec)
  (p : stpred) (g : acpred) (r : reg),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s
    -> exists p'', exists g'',
      OCHSpec.lookup OPsi (gSR s r) <* rho_scap, [p'' & g''] *>
      /\ optp p'' (nextS (jr r) s)
      /\ (forall s', 
        optg g'' (nextS (jr r) s) s' -> g s s'))
  -> wfis D << {[ <* rho_scap, [ p & g] *> | D ]} /OPsi>> 
  (ione (jr r)).
Proof.
  intros D OPsi p g r Hd Himp.
  apply wfis_jr.  
  unfold lift, itpscap, BigInterp.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Himp with (Himp s Hp).
  destruct Himp as [p'' [g'' Himp]].
  exists <* rho_scap, [p'' & g''] *>.
  destruct Himp as [Hin Himp].
  split_l using 
    (apply (OCHSpec.lookup_sub_lookup Hin Hsub)).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  destruct Himp as [Hp'' Himp].
  split_l.

  destruct Hk as [n Hk].
  exists n.
  destruct n.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
Qed.

Lemma scap_addu : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rd rs rt : reg) (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> optp p' (nextS (addu rd rs rt) s)
      /\ (forall s', 
        optg g' (nextS (addu rd rs rt) s) s' -> g s s'))
  -> wfis D <<[| p' & g'|] / OPsi>> is
  -> wfis D <<[| p & g|] / OPsi>> 
              (iseq (addu rd rs rt) is).
Proof.
  intros D OPsi p p' g g' rd rs rt is Hd Himp His.
  apply wfis_addu. 
  unfold lift, itpscap.
  exists (<<[|p' & g'|] / OPsi>>).
  split_r.
  intros Psi s [Hsub [Hp Hk]].
  substH Himp with (Himp s Hp).
  destruct (nextS (addu rd rs rt) s) as [s0 | ];
    [idtac | tauto].
  unfold lift, itpscap.
  split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_addu_2 : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rd rs rt : reg) (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> optp p' (nextS (addu rd rs rt) s)
      /\ (forall s', 
        optg g' (nextS (addu rd rs rt) s) s' -> g s s'))
  -> wfis D << {[ <* rho_scap, [ p' & g'] *> | D]} / OPsi>> is
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
              (iseq (addu rd rs rt) is).
Proof.
  intros D OPsi p p' g g' rd rs rt is Hd Himp His.
  apply wfis_addu. 
  unfold lift, itpscap.
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  split_r.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Himp with (Himp s Hp).
  destruct (nextS (addu rd rs rt) s) as [s0 | ];
    [idtac | tauto].
  unfold lift, itpscap.
  split_l.
  simpl; exists itpscap; split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_addiu : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rd rs : reg) w (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> optp p' (nextS (addiu rd rs w) s)
      /\ (forall s', 
        optg g' (nextS (addiu rd rs w) s) s' -> g s s'))
  -> wfis D <<[| p' & g'|] / OPsi>> is
  -> wfis D <<[| p & g|] / OPsi>>
              (iseq (addiu rd rs w) is).
Proof.
  intros D OPsi p p' g g' rd rs w is Hd Himp His.
  apply wfis_addiu. 
  unfold lift, itpscap.
  exists (<<[|p' & g'|] / OPsi>>).
  split_r.
  intros Psi s [Hsub [Hp Hk]].
  substH Himp with (Himp s Hp).
  destruct (nextS (addiu rd rs w) s) as [s0 | ];
    [idtac | tauto].
  unfold lift, itpscap.
  split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_addiu_2 : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rd rs : reg) w (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> optp p' (nextS (addiu rd rs w) s)
      /\ (forall s', 
        optg g' (nextS (addiu rd rs w) s) s' -> g s s'))
  -> wfis D << {[ <* rho_scap, [ p' & g'] *> | D]} / OPsi>> is
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
              (iseq (addiu rd rs w) is).
Proof.
  intros D OPsi p p' g g' rd rs w is Hd Himp His.
  apply wfis_addiu. 
  unfold lift, itpscap.
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  split_r.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Himp with (Himp s Hp).
  destruct (nextS (addiu rd rs w) s) as [s0 | ];
    [idtac | tauto].
  unfold lift, itpscap.
  split_l.
  simpl; exists itpscap; split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_subu : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rd rs rt : reg) (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> optp p' (nextS (subu rd rs rt) s)
      /\ (forall s', 
        optg g' (nextS (subu rd rs rt) s) s' -> g s s'))
  -> wfis D <<[| p' & g'|] / OPsi>> is
  -> wfis D <<[| p & g|] / OPsi>> 
              (iseq (subu rd rs rt) is).
Proof.
  intros D OPsi p p' g g' rd rs rt is Hd Himp His.
  apply wfis_subu. 
  unfold lift, itpscap.
  exists (<<[|p' & g'|] / OPsi>>).
  split_r.
  intros Psi s [Hsub [Hp Hk]].
  substH Himp with (Himp s Hp).
  destruct (nextS (subu rd rs rt) s) as [s0 | ];
    [idtac | tauto].
  unfold lift, itpscap.
  split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_subu_2 : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rd rs rt : reg) (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> optp p' (nextS (subu rd rs rt) s)
      /\ (forall s', 
        optg g' (nextS (subu rd rs rt) s) s' -> g s s'))
  -> wfis D << {[ <* rho_scap, [ p' & g'] *> | D]} / OPsi>> is
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
              (iseq (subu rd rs rt) is).
Proof.
  intros D OPsi p p' g g' rd rs rt is Hd Himp His.
  apply wfis_subu. 
  unfold lift, itpscap.
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  split_r.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Himp with (Himp s Hp).
  destruct (nextS (subu rd rs rt) s) as [s0 | ];
    [idtac | tauto].
  unfold lift, itpscap.
  split_l.
  simpl; exists itpscap; split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_subiu : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rd rs : reg) w (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> optp p' (nextS (subiu rd rs w) s)
      /\ (forall s', 
        optg g' (nextS (subiu rd rs w) s) s' -> g s s'))
  -> wfis D <<[| p' & g'|] / OPsi>> is
  -> wfis D <<[| p & g|] / OPsi>> 
              (iseq (subiu rd rs w) is).
Proof.
  intros D OPsi p p' g g' rd rs w is Hd Himp His.
  apply wfis_subiu. 
  unfold lift, itpscap.
  exists (<<[|p' & g'|] / OPsi>>).
  split_r.
  intros Psi s [Hsub [Hp Hk]].
  substH Himp with (Himp s Hp).
  destruct (nextS (subiu rd rs w) s) as [s0 | ];
    [idtac | tauto].
  unfold lift, itpscap.
  split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_subiu_2 : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rd rs : reg) w (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> optp p' (nextS (subiu rd rs w) s)
      /\ (forall s', 
        optg g' (nextS (subiu rd rs w) s) s' -> g s s'))
  -> wfis D << {[ <* rho_scap, [ p' & g'] *> | D]} / OPsi>> is
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
              (iseq (subiu rd rs w) is).
Proof.
  intros D OPsi p p' g g' rd rs w is Hd Himp His.
  apply wfis_subiu. 
  unfold lift, itpscap.
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  split_r.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Himp with (Himp s Hp).
  destruct (nextS (subiu rd rs w) s) as [s0 | ];
    [idtac | tauto].
  unfold lift, itpscap.
  split_l.
  simpl; exists itpscap; split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_lw : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rt rs : reg) w (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> optp p' (nextS (lw rt w rs) s)
      /\ (forall s', 
        optg g' (nextS (lw rt w rs) s) s' -> g s s'))
  -> wfis D <<[| p' & g'|] / OPsi>> is
  -> wfis D <<[| p & g|] / OPsi>> 
              (iseq (lw rt w rs) is).
Proof.
  intros D OPsi p p' g g' rt rs w is Hd Himp His.
  apply wfis_lw. 
  unfold lift, itpscap.
  exists (<<[|p' & g'|] / OPsi>>).
  split_r.
  intros Psi s [Hsub [Hp Hk]].
  substH Himp with (Himp s Hp).
  destruct (nextS (lw rt w rs) s) as [s0 | ];
    [idtac | tauto].
  unfold lift, itpscap.
  split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_lw_2 : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rt rs : reg) w (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> optp p' (nextS (lw rt w rs) s)
      /\ (forall s', 
        optg g' (nextS (lw rt w rs) s) s' -> g s s'))
  -> wfis D << {[ <* rho_scap, [ p' & g'] *> | D]} / OPsi>> is
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
              (iseq (lw rt w rs) is).
Proof.
  intros D OPsi p p' g g' rt rs w is Hd Himp His.
  apply wfis_lw. 
  unfold lift, itpscap.
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  split_r.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Himp with (Himp s Hp).
  destruct (nextS (lw rt w rs) s) as [s0 | ];
    [idtac | tauto].
  unfold lift, itpscap.
  split_l.
  simpl; exists itpscap; split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_sw : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rt rs : reg) w (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> optp p' (nextS (sw rt w rs) s)
      /\ (forall s', 
        optg g' (nextS (sw rt w rs) s) s' -> g s s'))
  -> wfis D <<[| p' & g'|] / OPsi>> is
  -> wfis D <<[| p & g|] / OPsi>> 
              (iseq (sw rt w rs) is).
Proof.
  intros D OPsi p p' g g' rt rs w is Hd Himp His.
  apply wfis_sw. 
  unfold lift, itpscap.
  exists (<<[|p' & g'|] / OPsi>>).
  split_r.
  intros Psi s [Hsub [Hp Hk]].
  substH Himp with (Himp s Hp).
  destruct (nextS (sw rt w rs) s) as [s0 | ];
    [idtac | tauto].
  unfold lift, itpscap.
  split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_sw_2 : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rt rs : reg) w (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> optp p' (nextS (sw rt w rs) s)
      /\ (forall s', 
        optg g' (nextS (sw rt w rs) s) s' -> g s s'))
  -> wfis D << {[ <* rho_scap, [ p' & g'] *> | D]} / OPsi>> is
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
              (iseq (sw rt w rs) is).
Proof.
  intros D OPsi p p' g g' rt rs w is Hd Himp His.
  apply wfis_sw. 
  unfold lift, itpscap.
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  split_r.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Himp with (Himp s Hp).
  destruct (nextS (sw rt w rs) s) as [s0 | ];
    [idtac | tauto].
  unfold lift, itpscap.
  split_l.
  simpl; exists itpscap; split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_blt : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rs rt : reg) (f : label) (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> if blt_nat (gSR s rs) (gSR s rt) 
      then 
        (exists p'', exists g'',
        OCHSpec.lookup OPsi f <* rho_scap, [p'' & g''] *>
        /\ optp p'' (nextS (blt rs rt f) s)
        /\ forall s', 
          optg g'' (nextS (blt rs rt f) s) s' -> g s s')
      else
        optp p' (nextS (blt rs rt f) s)
        /\ forall s',
          optg g' (nextS (blt rs rt f) s) s' -> g s s')
  -> wfis D <<[| p' & g'|] / OPsi>> is
  -> wfis D <<[| p & g|] / OPsi>> (iseq (blt rs rt f) is).
Proof.
  intros D OPsi p p' g g' rs rt f is Hd Himp His.
  apply wfis_blt. 
  unfold lift, itpscap.
  exists <<[|p' & g'|] /OPsi>>.
  split_r.
  intros Psi s [Hsub [Hp Hk]].
  substH Himp with (Himp s Hp).
  destruct (nextS (blt rs rt f) s) as [s0 | ].
  unfold lift, itpscap.

  bnat_blt_dec (gSR s rs) (gSR s rt).
  (* case rs < rt *)
  destruct Himp as [p'' [g'' Himp]].
  exists (<* rho_scap, [p'' & g''] *>).
  destruct Himp as [Hin Himp].
  split_l using 
    (apply (OCHSpec.lookup_sub_lookup Hin Hsub)).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  destruct Himp as [Hp'' Himp].
  split_l.

  destruct Hk as [n Hk].
  exists n.
  destruct n.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.

  (* case rs >= rt *)
  split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.

  (* .... *)
  bnat_blt_dec (gSR s rs) (gSR s rt).
  destruct Himp as [p'' [g'' [H1 [Hf H3]]]].
  destruct Hf.
  destruct Himp as [Hf _].
  destruct Hf.
Qed.

Lemma scap_blt_2 : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rs rt : reg) (f : label) (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> if blt_nat (gSR s rs) (gSR s rt) 
      then 
        (exists p'', exists g'',
        OCHSpec.lookup OPsi f <* rho_scap, [p'' & g''] *>
        /\ optp p'' (nextS (blt rs rt f) s)
        /\ forall s', 
          optg g'' (nextS (blt rs rt f) s) s' -> g s s')
      else
        optp p' (nextS (blt rs rt f) s)
        /\ forall s',
          optg g' (nextS (blt rs rt f) s) s' -> g s s')
  -> wfis D << {[ <* rho_scap, [ p' & g'] *> | D]} / OPsi>> is
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
            (iseq (blt rs rt f) is).
Proof.
  intros D OPsi p p' g g' rs rt f is Hd Himp His.
  apply wfis_blt. 
  unfold lift, itpscap.
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  split_r.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Himp with (Himp s Hp).
  destruct (nextS (blt rs rt f) s) as [s0 | ].
  unfold lift, itpscap.

  bnat_blt_dec (gSR s rs) (gSR s rt).
  (* case rs < rt *)
  destruct Himp as [p'' [g'' Himp]].
  exists (<* rho_scap, [p'' & g''] *>).
  destruct Himp as [Hin Himp].
  split_l using 
    (apply (OCHSpec.lookup_sub_lookup Hin Hsub)).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  destruct Himp as [Hp'' Himp].
  split_l.

  destruct Hk as [n Hk].
  exists n.
  destruct n.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
  
  (* case rs >= rt *)
  split_l.
  simpl; exists itpscap; split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.

  (* .... *)
  bnat_blt_dec (gSR s rs) (gSR s rt).
  destruct Himp as [p'' [g'' [H1 [Hf H3]]]].
  destruct Hf.
  destruct Himp as [Hf _].
  destruct Hf.
Qed.

Lemma scap_bgtz : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rs : reg) (f : label) (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> if blt_nat O (gSR s rs) 
      then 
        (exists p'', exists g'',
        OCHSpec.lookup OPsi f <* rho_scap, [p'' & g''] *>
        /\ optp p'' (nextS (bgtz rs f) s)
        /\ forall s', 
          optg g'' (nextS (bgtz rs f) s) s' -> g s s')
      else
        optp p' (nextS (bgtz rs f) s)
        /\ forall s',
          optg g' (nextS (bgtz rs f) s) s' -> g s s')
  -> wfis D <<[| p' & g'|] / OPsi>> is
  -> wfis D <<[| p & g|] / OPsi>> (iseq (bgtz rs f) is).
Proof.
  intros D OPsi p p' g g' rs f is Hd Himp His.
  apply wfis_bgtz. 
  unfold lift, itpscap.
  exists <<[|p' & g'|] /OPsi>>.
  split_r.
  intros Psi s [Hsub [Hp Hk]].
  substH Himp with (Himp s Hp).
  destruct (nextS (bgtz rs f) s) as [s0 | ].
  unfold lift, itpscap.

  bnat_blt_dec O (gSR s rs).
  (* case rs < rt *)
  destruct Himp as [p'' [g'' Himp]].
  exists (<* rho_scap, [p'' & g''] *>).
  destruct Himp as [Hin Himp].
  split_l using 
    (apply (OCHSpec.lookup_sub_lookup Hin Hsub)).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  destruct Himp as [Hp'' Himp].
  split_l.

  destruct Hk as [n Hk].
  exists n.
  destruct n.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.

  (* case rs >= rt *)
  split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.

  (* .... *)
  bnat_blt_dec O (gSR s rs).
  destruct Himp as [p'' [g'' [H1 [Hf H3]]]].
  destruct Hf.
  destruct Himp as [Hf _].
  destruct Hf.
Qed.

Lemma scap_bgtz_2 : forall (D : idict) (OPsi : chspec)
  (p p' : stpred) (g g' : acpred) 
  (rs : reg) (f : label) (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> if blt_nat O (gSR s rs) 
      then 
        (exists p'', exists g'',
        OCHSpec.lookup OPsi f <* rho_scap, [p'' & g''] *>
        /\ optp p'' (nextS (bgtz rs f) s)
        /\ forall s', 
          optg g'' (nextS (bgtz rs f) s) s' -> g s s')
      else
        optp p' (nextS (bgtz rs f) s)
        /\ forall s',
          optg g' (nextS (bgtz rs f) s) s' -> g s s')
  -> wfis D << {[ <* rho_scap, [ p' & g'] *> | D]} / OPsi>> is
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
           (iseq (bgtz rs f) is).
Proof.
  intros D OPsi p p' g g' rs f is Hd Himp His.
  apply wfis_bgtz. 
  unfold lift, itpscap.
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  split_r.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Himp with (Himp s Hp).
  destruct (nextS (bgtz rs f) s) as [s0 | ].
  unfold lift, itpscap.

  bnat_blt_dec O (gSR s rs).
  (* case rs < rt *)
  destruct Himp as [p'' [g'' Himp]].
  exists (<* rho_scap, [p'' & g''] *>).
  destruct Himp as [Hin Himp].
  split_l using 
    (apply (OCHSpec.lookup_sub_lookup Hin Hsub)).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  destruct Himp as [Hp'' Himp].
  split_l.

  destruct Hk as [n Hk].
  exists n.
  destruct n.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.

  (* case rs >= rt *)
  split_l.
  simpl; exists itpscap; split_l.
  destruct Himp as [Hp' Hg'].
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.

  (* .... *)
  bnat_blt_dec O (gSR s rs).
  destruct Himp as [p'' [g'' [H1 [Hf H3]]]].
  destruct Hf.
  destruct Himp as [Hf _].
  destruct Hf.
Qed.


Lemma scap_wfis_ocap_wfis :
  forall D OPsi is psi p g,
    DSpec.sub D0 D
    -> DSpec.lookup D rho_scap <@itpscap@>
    -> scap_chspec_lift psi OPsi
    -> scap_wfis psi p g is
    -> ocap_wfis D <<[| p & g|] / OPsi>> is.
Proof.
  intros D OPsi is.
  induction is.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    intros psi p g Hd Hin Hlift Hscap;
      try inversion_clear Hscap.

  apply scap_tcall; trivial.
  intros s Hp.
  destruct (H s Hp) as [p'' [g'' [Hin' H']]].
  exists p''; exists g''.
  split_r.
  apply ((proj1 Hlift) _ _ _ Hin').

  apply scap_call; trivial.
  intros s Hp.
  substH H with (H s Hp). 
  destruct H as [p'' [g'' [Hin' H']]].
  exists p''; exists g''.
  split_l with ((proj1 Hlift) _ _ _ Hin').
  destruct H' as [Hp'' H].
  split_l.
  intros s' Hg''.
  substH H with (H s' Hg'').
  destruct H as [p''' [g''' [Hin''' [Hp''' Hg''']]]].
  exists p'''; exists g'''.
  split_l with ((proj1 Hlift) _ _ _ Hin''').
  split_lr.

  apply scap_jr; trivial.
  intros s Hp.
  destruct (H s Hp) as [p'' [g'' [Hin' H']]].
  exists p''; exists g''.
  split_r.
  apply ((proj1 Hlift) _ _ _ Hin').

  apply scap_ret; trivial.
  
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    intros psi p g Hd Hin Hlift Hscap;
      try inversion_clear Hscap.
  
  apply scap_addu with p' g'; trivial.
  apply IHis with psi; trivial.

  apply scap_addiu with p' g'; trivial.
  apply IHis with psi; trivial.

  apply scap_subu with p' g'; trivial.
  apply IHis with psi; trivial.

  apply scap_subiu with p' g'; trivial.
  apply IHis with psi; trivial.
  
  apply scap_lw with p' g'; trivial.
  apply IHis with psi; trivial.
  
  apply scap_sw with p' g'; trivial.
  apply IHis with psi; trivial.
  
  apply scap_blt with p' g'; trivial.
  intros s Hp.
  substH H with (H s Hp).
  unfoldgs.
  bnat_blt_dec (gSR s rs) (gSR s rt).
  destruct H as [p'' [g'' [Hin'' H]]].
    exists p''; exists g''.
    split_l with ((proj1 Hlift) _ _ _ Hin'').
    trivial.
  trivial.
  apply IHis with psi; trivial.

  apply scap_bgtz with p' g'; trivial.
  intros s Hp.
  substH H with (H s Hp).
  unfoldgs.
  bnat_blt_dec O (gSR s rs).
  destruct H as [p'' [g'' [Hin'' H]]].
    exists p''; exists g''.
    split_l with ((proj1 Hlift) _ _ _ Hin'').
    trivial.
  trivial.
  apply IHis with psi; trivial.
Qed.

Lemma scap_wfis_ocap_wfis_2 :
  forall D OPsi is psi p g,
    DSpec.sub D0 D
    -> DSpec.lookup D rho_scap <@itpscap@>
    -> scap_chspec_lift psi OPsi
    -> scap_wfis psi p g is
    -> ocap_wfis D << {[ <* rho_scap, [p & g] *> | D]} / OPsi>> is.
Proof.
  intros D OPsi is.
  induction is.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    intros psi p g Hd Hin Hlift Hscap;
      try inversion_clear Hscap.

  apply scap_tcall_2; trivial.
  intros s Hp.
  destruct (H s Hp) as [p'' [g'' [Hin' H']]].
  exists p''; exists g''.
  split_r.
  apply ((proj1 Hlift) _ _ _ Hin').

  apply scap_call_2; trivial.
  intros s Hp.
  substH H with (H s Hp). 
  destruct H as [p'' [g'' [Hin' H']]].
  exists p''; exists g''.
  split_l with ((proj1 Hlift) _ _ _ Hin').
  destruct H' as [Hp'' H].
  split_l.
  intros s' Hg''.
  substH H with (H s' Hg'').
  destruct H as [p''' [g''' [Hin''' [Hp''' Hg''']]]].
  exists p'''; exists g'''.
  split_l with ((proj1 Hlift) _ _ _ Hin''').
  split_lr.

  apply scap_jr_2; trivial.
  intros s Hp.
  destruct (H s Hp) as [p'' [g'' [Hin' H']]].
  exists p''; exists g''.
  split_r.
  apply ((proj1 Hlift) _ _ _ Hin').

  apply scap_ret_2; trivial.
  
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    intros psi p g Hd Hin Hlift Hscap;
      try inversion_clear Hscap.
  
  apply scap_addu_2 with p' g'; trivial.
  apply IHis with psi; trivial.

  apply scap_addiu_2 with p' g'; trivial.
  apply IHis with psi; trivial.

  apply scap_subu_2 with p' g'; trivial.
  apply IHis with psi; trivial.

  apply scap_subiu_2 with p' g'; trivial.
  apply IHis with psi; trivial.
  
  apply scap_lw_2 with p' g'; trivial.
  apply IHis with psi; trivial.
  
  apply scap_sw_2 with p' g'; trivial.
  apply IHis with psi; trivial.
  
  apply scap_blt_2 with p' g'; trivial.
  intros s Hp.
  substH H with (H s Hp).
  unfoldgs.
  bnat_blt_dec (gSR s rs) (gSR s rt).
  destruct H as [p'' [g'' [Hin'' H]]].
    exists p''; exists g''.
    split_l with ((proj1 Hlift) _ _ _ Hin'').
    trivial.
  trivial.
  apply IHis with psi; trivial.

  apply scap_bgtz_2 with p' g'; trivial.
  intros s Hp.
  substH H with (H s Hp).
  unfoldgs.
  bnat_blt_dec O (gSR s rs).
  destruct H as [p'' [g'' [Hin'' H]]].
    exists p''; exists g''.
    split_l with ((proj1 Hlift) _ _ _ Hin'').
    trivial.
  trivial.
  apply IHis with psi; trivial.
Qed.

Lemma D_scap_describe_lift_psi :
  forall OPsi psi,
    scap_chspec_lift psi OPsi
    -> describe D_scap OPsi.
Proof.
  intros OPsi psi.
  unfold scap_chspec_lift.
  intros H f rho lty theta Hin.
  destruct H as [H1 H2].
  destruct (H2 f <*rho,theta*> Hin) as [p [g [Hpi Hin']]].
  injection Hpi.
  intros Heq1 Heq2 Heq3.
  subst.
  exists itpscap.
  apply DSpec.get_lookup.
  unfold D_scap, DSpec.get, DSpecDef.get.
  bnat_rew (beq_nat rho_scap rho_scap = true).
  trivial.
Qed.

Lemma scap_wfch_ocap_wfch :
  forall ch psi psi' OPsi OPsi',
    scap_chspec_lift psi OPsi
    -> scap_chspec_lift psi' OPsi'
    -> scap_wfch psi psi' ch
    -> ocap_wfch (DSpec.merge D0 D_scap) OPsi OPsi' ch.
Proof.
  intros ch psi psi' OPsi OPsi'.
  intros Hpsi Hpsi' Hch.
  unfold WF_C.
  split.
    split.
      assert (H : D_scap ## OPsi).
      eapply D_scap_describe_lift_psi; eauto.
      apply describe_sub with D_scap; trivial.
      apply D_scap_sub_D1.
    assert (H : D_scap ## OPsi').
    eapply D_scap_describe_lift_psi; eauto.
    apply describe_sub with D_scap; trivial.
    apply D_scap_sub_D1.
  intros f pi Hin.
  unfold scap_chspec_lift in Hpsi'.
  unfold wfch in Hch.
  destruct Hpsi' as [_ Hpsi'].
  destruct (Hpsi' f pi Hin) as [p [g [Hpi Hin']]].
  subst pi.
  
  destruct (Hch f p g (CHSpec.lookup_get Hin')) as [is [Hgis Hwfis]].
  exists is; split_l.
  apply scap_wfis_ocap_wfis_2 with psi.
  apply D0_sub_D1.
  apply DSpec.lookup_merge_intro.
  right.
  apply DSpec.get_lookup.
  unfold D_scap, DSpec.get, DSpecDef.get.
  bnat_rew (beq_nat rho_scap rho_scap = true).
  trivial.
  apply DSpec.disj_compat.
  apply DSpec.disj_sym.
  apply D_scap_disj_D0.
  trivial.
  trivial.
Qed.

Theorem scap_sound :
  forall ch psi psi' OPsi OPsi',
    scap_chspec_lift psi OPsi
    -> scap_chspec_lift psi' OPsi'
    -> scap_wfch psi psi' ch
    -> ocap_wfch (DSpec.merge D0 D_scap) OPsi OPsi' ch.
Proof scap_wfch_ocap_wfch.

End ScapOcap.
