(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(*                                                             *)
(*  Additional TAL lemmas                                      *)
(*                                                             *)
(***************************************************************)

(* $Id: sep_tal.v,v 1.2 2006/07/15 02:56:41 ssg05 Exp $ *)

Require Import machine.
Require Import tal0.
Require Import seplogic.
Require Import natpm.
Require Import extac.

Open Local Scope seplogic_scope.
Open Local Scope tal_scope.

Lemma sep_eq : 
  forall h h' l v,
    (h' ||- (fun H : heap => H = h) ** l |-> v)
    -> h' = (Heap.set h l v).
Proof.
  intros h h' l v H.
  unfold hpred_true, sconj in H.
  destruct H as [ha [hb [J [Hha Hhb]]]].
  unfold hpred_true in Hha.
  unfold hpred_true in Hhb.
  unfold cont in Hhb.
  rewrite Hha in J.
  rewrite Hhb in J.
  rewrite (Heap.join_sig_r J).
  trivial.
Qed.

Lemma WF_H_weaken_H : forall psi phi h h1 h2,
  Heap.join h1 h2 h
  -> WF_H psi phi h1 
  -> WF_H psi phi h.
Proof.
  intros psi phi h h1 h2 Hj Hh1.
  unfold WF_H.
  intros l ct Hin.
  substH Hh1 with (Hh1 l ct Hin).
  destruct Hh1 as [v [Hvin Hctoke]].
  exists v.
  split_l.
  eapply Heap.join_lookup_lookup_l; eauto.
  trivial.
Qed.

Lemma heap_extend_sig_WF_H :
  forall h h' l v talpsi Phi,
    (h' ||- (fun H : heap => H = h) ** l |-> v ** top)
    -> WF_H talpsi Phi h
    -> WF_H talpsi
    (DHSpec.set Phi l (tal0.junk, tal0.valty_int))
    h'.
Proof.
  intros h h' l v psi phi Hsep Hh.
  assert (Hsep' : h' ||- ((fun H : heap => H = h) 
    ** l |-> v) ** top).
  simplsep 3 3. clear Hsep.
  destruct Hsep' as [h1 [h2 [J [Hh1 Hh2]]]].
  apply WF_H_weaken_H with h1 h2; trivial.
  rewrite (sep_eq h h1 l v Hh1).
  apply heap_extend_WF_H; trivial.
  destruct Hh1 as [h1a [h1b [h1J [Hh1a Hh1b]]]].
  unfold hpred_true in Hh1a.
  subst h1a.
  unfold WF_H in Hh.
  destruct (DHSpec.get_dec phi l) as [Hsome | Hnone].
  destruct Hsome as [b Hsome].
  substH Hsome with (DHSpec.get_lookup Hsome).
  destruct (Hh l b Hsome) as [vv [Hvvin Hct]].
  substH Hh1b with (cont_get_eq Hh1b).
  substH Hvvin with (Heap.lookup_get Hvvin).
  unfold Heap.join in h1J.
  substH h1J with (h1J l).
  rewrite Hh1b in h1J.
  rewrite Hvvin in h1J.
  destruct h1J.
  trivial.
  constructor.
Qed.

Lemma heap_extend_pair_WF_H :
  forall h h' l1 l2 v1 v2 talpsi Phi,
    (h' ||- (fun H : heap => H = h) 
      ** l1 |-> v1 ** l2 |-> v2 ** top)
    -> WF_H talpsi Phi h
    -> WF_H talpsi
    (DHSpec.set 
      (DHSpec.set Phi l1 (junk, valty_int))
      l2 (junk, valty_int)) h'.
Proof.
  intros h h' l1 l2 v1 v2 psi phi.
  intros Hsep Hh.
  assert (H0 : h' ||- ((fun H : heap => H = h) 
    ** l1 |-> v1) ** l2 |-> v2 ** top).
  simplsep 4 4.
  destruct H0 as [h1' [h2' [Jh' [Hh1' Hh2']]]].
  assert (H : WF_H psi (DHSpec.set phi l1 (junk,int)) h1').
  assert (H1 : h1' ||- ((fun H : heap => H = h) 
    ** l1 |-> v1) ** top).
  apply top_intro_right; trivial.
  assert (H2 : h1' ||- (fun H : heap => H = h) 
    ** l1 |-> v1 ** top).
  simplsep 3 3.
  eapply heap_extend_sig_WF_H; eauto.
  assert (H3 : h' ||- (fun H : heap => H = h1') 
    ** l2 |-> v2 ** top).
  simplsep 3 4.
  destruct Hsep as [hb [hc [J1 [Hhb Hhc]]]].
  destruct Hh1' as [ha [hc' [J2 [Hha Hhc']]]].
  unfold hpred_true in * .
  subst ha hb.
  unfold cont in Hhc', Hhc.
  rewrite <- Hhc' in Hhc.
  clear Hhc'.
  subst hc'.
  eapply Heap.join_eq; eauto.
  eapply heap_extend_sig_WF_H; eauto.
Qed. 

Lemma heap_extend_pair_regtype :
  forall h addr v1 v2 talpsi talphi,
    (h ||- addr |-> v1 ** S addr |-> v2 ** top)
    -> valok talpsi 
    (DHSpec.set 
      (DHSpec.set talphi
        addr (junk, int))
        (S addr) (junk, int)) addr
    [<(junk, int), (junk, int)>].
Proof.
  intros h addr v1 v2 psi phi Hsep.
  apply valok_tup.
  intros i Hiin.
  exists (junk, int).
  exists (junk, int).
  destruct Hiin as [ct Hlook].
  destruct i.
  simpl in Hlook |- * .
  split_l.
  split_r using (constructor; trivial).
  simplpm.
  DHSpec.simplgs.
  trivial.
  omega.
  destruct i.
  simpl in Hlook |- * .
  split_l.
  split_r using (constructor; trivial).
  simplpm.
  apply DHSpec.get_lookup.
  DHSpec.simplgs.
  trivial.
  destruct i.
  simpl in Hlook |- * .
  destruct Hlook.
  simpl in Hlook |- * .
  destruct Hlook.
Qed.