(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* TAL over OCAP                                               *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: tal_ocap.v,v 1.4 2006/07/15 02:56:41 ssg05 Exp $ *)

Require Import machine.
Require tal0.
Require Import ocap.

Require Import extac.
Require Import deptac.

Open Local Scope tm_scope.
Open Local Scope ocap_scope.
Open Local Scope tal_scope.

Module Type TALPARAM.

  Parameter rho_tal : nat.
  Parameter open_r : assert.
  Axiom open_r_only_assert_heap :
    forall Psi H R R' pc pc',
      open_r Psi <.H, R, pc.> 
      -> open_r Psi <.H, R', pc'.>.
End TALPARAM.

Module TalOcap (Params : TALPARAM).

Import Params.

Inductive Ltal : Type := 
  | mk_Ltal : tal0.rfTy -> Ltal.

(* Notation Ltal := tal0.codeTy. *)
(* Notation mk_Ltal := tal0.valty_code. *)
Notation tal_chspec := tal0.chspec.
Notation tal_wfis := tal0.wfis.
Notation ocap_wfis := ocap.wfis.
Notation tal_wfch := tal0.WF_C.
Notation ocap_wfch := ocap.WF_C.

Module TalCHSpec := tal0.CHSpec.

(*************************************************** *)

Definition tal_chspec_lift (psi : tal_chspec) (OPsi : chspec) : Prop :=
  (forall f (Y : tal0.rfTy),
    TalCHSpec.lookup psi f Y
    -> OCHSpec.lookup OPsi f <* rho_tal, (mk_Ltal Y) *>)
  /\ (forall f pi, 
    OCHSpec.lookup OPsi f pi
    -> exists Y, 
      pi = <* rho_tal, mk_Ltal Y *>
      /\ TalCHSpec.lookup psi f Y).

Definition tal_chspec_sink (OPsi : chspec) (psi : tal_chspec) : Prop :=
  forall f (Y : tal0.rfTy),
    (OCHSpec.lookup OPsi f <* rho_tal, (mk_Ltal Y) *>
    -> TalCHSpec.lookup psi f Y)
    /\ (TalCHSpec.lookup psi f Y
      -> OCHSpec.lookup OPsi f <* rho_tal, (mk_Ltal Y) *>) .

Lemma lift_lookup : forall psi OPsi f Y,
  tal_chspec_lift psi OPsi
  -> TalCHSpec.lookup psi f Y
  -> OCHSpec.lookup OPsi f <*rho_tal, mk_Ltal Y*> .
Proof.
  intros psi OPsi f Y.
  unfold tal_chspec_lift.
  intros [H1 H2] Hin.
  apply H1; trivial.
Qed.

Lemma sink_lookup : forall psi OPsi f Y,
  tal_chspec_sink OPsi psi
  -> TalCHSpec.lookup psi f Y
  -> OCHSpec.lookup OPsi f <*rho_tal, mk_Ltal Y*> .
Proof.
  intros psi OPsi f Y.
  unfold tal_chspec_sink.
  intros H H1.
  substH H with (H f Y).
  destruct H.
  auto.
Qed.


Lemma lift_sub : forall psi psi' OPsi Psi f Y,
  tal_chspec_lift psi OPsi
  -> OCHSpec.sub OPsi Psi
  -> TalCHSpec.lookup psi' f Y
  -> tal_chspec_sink Psi psi'
  -> OCHSpec.lookup Psi f <*rho_tal, mk_Ltal Y*>.
Proof.
  intros psi psi' OPsi Psi f Y.
  intros Hlift Hsub Hf Hsink.
  destruct (Hsink f Y) as [Hin1 Hin2].
  auto.
Qed.

Lemma sink_sub : forall psi psi' OPsi Psi f Y,
  tal_chspec_lift psi OPsi
  -> OCHSpec.sub OPsi Psi
  -> TalCHSpec.lookup psi' f Y
  -> tal_chspec_sink Psi psi'
  -> OCHSpec.lookup Psi f <*rho_tal, mk_Ltal Y*>.
Proof.
  intros psi psi' OPsi Psi f Y.
  intros Hlift Hsub Hf Hsink.
  destruct (Hsink f Y) as [Hin1 Hin2].
  auto.
Qed.

Lemma sink_sub_2 : forall psi psi' OPsi Psi f Y,
  tal_chspec_lift psi OPsi
  -> OCHSpec.sub OPsi Psi
  -> TalCHSpec.lookup psi f Y
  -> tal_chspec_sink Psi psi'
  -> TalCHSpec.lookup psi' f Y.
Proof.
  intros psi psi' OPsi Psi f Y.
  intros Hlift Hsub Hf Hsink.
  destruct (Hsink f Y) as [Hin1 Hin2].
  destruct Hlift.
  auto.
Qed.

(*************************************************** *)

Definition itptal (theta : Ltal) : assert := 
  match theta with 
    | mk_Ltal Y =>
  fun (OPsi : chspec) (s : state)  =>
    match s with 
      | <.H,R,pc.> =>
        exists H1, exists H2,
          Heap.join H1 H2 H
          /\ (exists psi, 
            tal_chspec_sink OPsi psi
            /\ tal0.WF_S psi Y <.H1,R,pc.>)
          /\ open_r OPsi <.H2,R,pc.>
    end
  end.

Definition D_tal : idict := fun id =>
  if (beq_nat id rho_tal) then SomeT <@ itptal @> else
    NoneT.

(*************************************************** *)

Lemma tal_wfis_ocap_wfis :
  forall D OPsi is psi Y,
    DSpec.lookup D rho_tal <@itptal@>
    -> tal_chspec_lift psi OPsi
    -> tal_wfis psi Y is
    -> ocap_wfis D <<(itptal (mk_Ltal Y)) / OPsi>> is.
Proof.
  intros D OPsi is.
  induction is.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    intros psi Y Hin Hlift Htal;
      generalize Hlift; intros [Hlift1 Hlift2];
      inversion_clear Htal as 
        [ p0 g0 Gamma'' f0 Hf Hsub
          | p0 g0 Gamma'' r0 Hr Hsub
          | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrsnr0 Hrs Hrt HI'
          | p0 g0 t r0 r1 r2 i0 Hrdnr0 Hrsr0 Hrt HI'
          | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrsnr0 Hrs HI'
          | p0 g0 Gamma'' r0 r1 w0 i0 Hrdnr0 Hrsr0 Hf HI'
          | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrs Hrt HI'
          | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrs HI'
          | p0 g0 o t r0 w0 r1 i0 Hrtnr0 Hrs Ht HI'
          | p0 g0 o ct t r0 w0 r1 i0 Hrsnr0 Hrs Hct Hsub Hrt HI'
          | p0 g0 Gamma'' r0 r1 f0 i0 Hf Hrs Hrt Hsub HI'
          | p0 g0 Gamma'' r0 f0 i0 Hf Hrs Hsub HI'
        ].

  (* case wfi_j *)
  apply wfis_j.
  unfold lift, itptal, BigInterp.
  intros Psi s [Hpsub Hwfs].
  exists <* rho_tal, (mk_Ltal Gamma'') *>.
  split_l.
  apply (OCHSpec.lookup_sub_lookup 
    (Hlift1 f Gamma'' Hf) Hpsub).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  exists itptal.
  split_l.
  simpl.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  inversion_clear Hsub.
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  eapply tal0.regfile_weakening; eauto.

  (* case wfis_jr *)
  apply wfis_jr.
  unfold lift, itptal, BigInterp.
  intros Psi s [Hpsub Hwfs].
  exists <* rho_tal, (mk_Ltal Gamma'') *>.
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  split_l.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  assert(Hx := tal0.reg_typing HR Hr).
  inversion_clear Hx.
  eapply sink_sub; eauto. 
  simpl.
  exists itptal.
  split_l.
  simpl.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  inversion_clear Hsub.
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  eapply tal0.regfile_weakening; eauto. 
  
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    intros psi Y Hin Hlift Htal;
      generalize Hlift; intros [Hlift1 Hlift2];
      inversion_clear Htal as
        [ p0 g0 Gamma'' f0 Hf Hsub
          | p0 g0 Gamma'' r0 Hr Hsub
          | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrsnr0 Hrs Hrt HI'
          | p0 g0 t r0 r1 r2 i0 Hrdnr0 Hrsr0 Hrt HI'
          | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrsnr0 Hrs HI'
          | p0 g0 Gamma'' r0 r1 w0 i0 Hrdnr0 Hrsr0 Hf HI'
          | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrs Hrt HI'
          | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrs HI'
          | p0 g0 o t r0 w0 r1 i0 Hrtnr0 Hrs Ht HI'
          | p0 g0 o ct t r0 w0 r1 i0 Hrsnr0 Hrs Hct Hsub Hrt HI'
          | p0 g0 Gamma'' r0 r1 f0 i0 Hf Hrs Hrt Hsub HI'
          | p0 g0 Gamma'' r0 f0 i0 Hf Hrs Hsub HI'
        ].

  (* case wfis_addu *)
  apply wfis_addu.
  unfold lift, itptal, BigInterp.
  exists << itptal (mk_Ltal (tal0.RfTy.upd Y rd 
    tal0.valty_int))/ OPsi >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  unfold lift.
  split_l.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  eapply tal0.regfile_update; eauto; 
    apply tal0.valok_int.

  (* case wfis_addu_mov *)
  apply wfis_addu.
  unfold lift, itptal, BigInterp.
  exists << itptal (mk_Ltal (tal0.RfTy.upd Y rd t))/ OPsi >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  unfold lift.
  split_l.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  breg_simpl; regfile.simplgs;  simpl.
  inversion_clear Hrt.
  eapply tal0.regfile_update; eauto.
  
  (* case wfis_addiu *)
  apply wfis_addiu.
  unfold lift, itptal, BigInterp.
  exists << itptal (mk_Ltal (tal0.RfTy.upd Y rd 
    tal0.valty_int))/ OPsi >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  unfold lift.
  split_l.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]]. 
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  breg_simpl; regfile.simplgs;  simpl.
  eapply tal0.regfile_update; eauto.
  apply tal0.valok_int; auto.

  (* case wfis_addiu_movi *)
  apply wfis_addiu.
  unfold lift, itptal, BigInterp.
  exists << itptal (mk_Ltal (tal0.RfTy.upd Y rd 
    (tal0.valty_code Gamma'')))/ OPsi >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  unfold lift.
  split_l.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  breg_simpl; regfile.simplgs;  simpl.
  eapply tal0.regfile_update; eauto.
  apply tal0.valok_lab; auto.
  eapply sink_sub_2; eauto.

  (* case wfis_subu *)
  apply wfis_subu.
  unfold lift, itptal, BigInterp.
  exists << itptal (mk_Ltal (tal0.RfTy.upd Y rd 
    tal0.valty_int))/ OPsi >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  unfold lift.
  split_l.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  eapply tal0.regfile_update; eauto; 
    apply tal0.valok_int.

  (* case wfis_subiu *)
  apply wfis_subiu.
  unfold lift, itptal, BigInterp.
  exists << itptal (mk_Ltal (tal0.RfTy.upd Y rd 
    tal0.valty_int))/ OPsi >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  unfold lift.
  split_l.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  eapply tal0.regfile_update; eauto.
  apply tal0.valok_int; auto.

  (* case wfis_lw *)
  apply wfis_lw.
  unfold lift, itptal, BigInterp.
  exists << itptal (mk_Ltal (tal0.RfTy.upd Y rt 
    t))/ OPsi >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  assert (Hval :=tal0.reg_typing HR Hrs).
  destruct (tal0.heapval_typing HH1 Ht Hval)
    as [v [Hvin Hv]].
  rewrite (Heap.join_lookup_lookup_l H12 Hvin).
  unfold lift.
  split_l.
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  eapply tal0.regfile_update; eauto.

  (* case wfis_sw *)
  apply wfis_sw.
  unfold lift, itptal, BigInterp.
  exists << itptal (mk_Ltal 
    (tal0.RfTy.upd Y rs
      (tal0.valty_tup 
        (tal0.TupTy.upd o i (tal0.init, t))))
    ) / OPsi >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  assert (Hval :=tal0.reg_typing HR Hrs).
  destruct (tal0.heapval_exists HH1 Hct Hval)
    as [v Hvin].
  rewrite (Heap.join_lookup_lookup_l H12 Hvin).
  unfold lift.
  split_l.
  exists (Heap.set H1 
    (Regfile.get rf rs + i) (Regfile.get rf rt)).
  exists H2.
  split_l.
  eapply Heap.join_set_l; eauto.
  eapply Heap.lookup_indom; eauto.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  destruct (tal0.heap_typing_reg_tying_ex_t
    HH1 HR Hrs Hct) as [ct' [Hct'in Hsub']].
  set (addr := (Regfile.get rf rs + i)) in * .
  set (Phi' := tal0.DHSpec.set Phi addr 
    (tal0.init, t)).
  exists psi'.
  split_l.
  exists Phi'.
    assert (Hx := tal0.DHSpec.lookup_canupd 
      (m:=Phi) (a:=addr) 
      (b:= ct') (tal0.init, t) Hct'in).
    assert (Hsubx := tal0.ceTy_sub_init_ct 
      t ct ct' Hsub' Hsub).
    split.
    refine (tal0.heap_update_WF_H HH1 Hsubx Hct'in Hx _).
    apply tal0.ceok_init.
    eapply tal0.reg_typing; eauto.
    eapply tal0.heap_update_WF_R; eauto.
    eapply tal0.heap_update_WF_R_reg; eauto.

  (* case wfis_blt *)
  apply wfis_blt.
  exists << itptal (mk_Ltal Y)/ OPsi >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  bnat_blt_dec (Regfile.get rf rs) (Regfile.get rf rt).
  exists <* rho_tal, (mk_Ltal Gamma'') *>.
    split_l.
    apply (OCHSpec.lookup_sub_lookup 
      (Hlift1 f Gamma'' Hf) Hpsub).
    simpl.
    rewrite Hblt.
    exists itptal.
    split_l.
    simpl.
    destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
    destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
    exists H1; exists H2.
    split_l.
    split_r using 
      (eapply open_r_only_assert_heap; eauto).
    exists psi'.
    split_l.
    exists Phi.
    split_l.
    inversion_clear Hsub.
    eapply tal0.regfile_weakening; eauto.
  simpl.
    rewrite Hnblt.
    unfold lift.
    split_l.
    destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
    destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
    exists H1; exists H2.
    split_l.
    split_r using 
      (eapply open_r_only_assert_heap; eauto).
    exists psi'.
    split_l.
    exists Phi.
    split_lr.

  (* case wfis_bgtz *)
  apply wfis_bgtz.
  exists << itptal (mk_Ltal Y)/ OPsi >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  bnat_blt_dec O (Regfile.get rf rs).
  exists <* rho_tal, (mk_Ltal Gamma'') *>.
    split_l.
    apply (OCHSpec.lookup_sub_lookup 
      (Hlift1 f Gamma'' Hf) Hpsub).
    unfold nextS.
    subst s.
    cbv iota beta.
    rewrite Hblt.
    simpl.
    exists itptal.
    split_l.
    simpl.
    destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
    destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
    exists H1; exists H2.
    split_l.
    split_r using 
      (eapply open_r_only_assert_heap; eauto).
    exists psi'.
    split_l.
    exists Phi.
    inversion_clear Hsub.
    split_l.
    eapply tal0.regfile_weakening; eauto.
  unfold nextS.
    subst s; cbv iota beta.
    rewrite Hnblt.
    unfold lift.
    split_l.
    destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
    destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
    exists H1; exists H2.
    split_l.
    split_r using 
      (eapply open_r_only_assert_heap; eauto).
    exists psi'.
    split_l.
    exists Phi.
    split_lr.
Qed.

Lemma tal_wfis_ocap_wfis_2 :
  forall D OPsi is psi Y,
    DSpec.lookup D rho_tal <@itptal@>
    -> tal_chspec_lift psi OPsi
    -> tal_wfis psi Y is
    -> ocap_wfis D <<{[ <*rho_tal, mk_Ltal Y *> | D ]} / OPsi>> is.
Proof.
  intros D OPsi is.
  simpl.
  induction is.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    intros psi Y Hin Hlift Htal;
      generalize Hlift; intros [Hlift1 Hlift2];
      inversion_clear Htal as 
        [ p0 g0 Gamma'' f0 Hf Hsub
          | p0 g0 Gamma'' r0 Hr Hsub
          | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrsnr0 Hrs Hrt HI'
          | p0 g0 t r0 r1 r2 i0 Hrdnr0 Hrsr0 Hrt HI'
          | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrsnr0 Hrs HI'
          | p0 g0 Gamma'' r0 r1 w0 i0 Hrdnr0 Hrsr0 Hf HI'
          | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrs Hrt HI'
          | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrs HI'
          | p0 g0 o t r0 w0 r1 i0 Hrtnr0 Hrs Ht HI'
          | p0 g0 o ct t r0 w0 r1 i0 Hrsnr0 Hrs Hct Hsub Hrt HI'
          | p0 g0 Gamma'' r0 r1 f0 i0 Hf Hrs Hrt Hsub HI'
          | p0 g0 Gamma'' r0 f0 i0 Hf Hrs Hsub HI'
        ].

  (* case wfi_j *)
  apply wfis_j.
  unfold lift, itptal, BigInterp.
  intros Psi s [Hpsub Hwfs].
  exists <* rho_tal, (mk_Ltal Gamma'') *>.
  split_l.
  apply (OCHSpec.lookup_sub_lookup 
    (Hlift1 f Gamma'' Hf) Hpsub).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  exists itptal.
  split_l.
  simpl.
  destruct Hwfs as [itp [Hin' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique Hin Hin').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear Hin' Hitpeq.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  inversion_clear Hsub.
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  eapply tal0.regfile_weakening; eauto.

  (* case wfis_jr *)
  apply wfis_jr.
  unfold lift, itptal, BigInterp.
  intros Psi s [Hpsub Hwfs].
  exists <* rho_tal, (mk_Ltal Gamma'') *>.
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  split_l.
  destruct Hwfs as [itp [Hin' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique Hin Hin').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear Hin' Hitpeq.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  assert(Hx := tal0.reg_typing HR Hr).
  inversion_clear Hx.
  eapply sink_sub; eauto. 
  simpl.
  exists itptal.
  split_l.
  simpl.
  destruct Hwfs as [itp [Hin' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique Hin Hin').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear Hin' Hitpeq.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  inversion_clear Hsub.
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  eapply tal0.regfile_weakening; eauto. 
  
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    intros psi Y Hin Hlift Htal;
      generalize Hlift; intros [Hlift1 Hlift2];
      inversion_clear Htal as
        [ p0 g0 Gamma'' f0 Hf Hsub
          | p0 g0 Gamma'' r0 Hr Hsub
          | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrsnr0 Hrs Hrt HI'
          | p0 g0 t r0 r1 r2 i0 Hrdnr0 Hrsr0 Hrt HI'
          | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrsnr0 Hrs HI'
          | p0 g0 Gamma'' r0 r1 w0 i0 Hrdnr0 Hrsr0 Hf HI'
          | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrs Hrt HI'
          | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrs HI'
          | p0 g0 o t r0 w0 r1 i0 Hrtnr0 Hrs Ht HI'
          | p0 g0 o ct t r0 w0 r1 i0 Hrsnr0 Hrs Hct Hsub Hrt HI'
          | p0 g0 Gamma'' r0 r1 f0 i0 Hf Hrs Hrt Hsub HI'
          | p0 g0 Gamma'' r0 f0 i0 Hf Hrs Hsub HI'
        ].

  (* case wfis_addu *)
  apply wfis_addu.
  unfold lift, itptal, BigInterp.
  exists <<
           (fun (psi0 : chspec) (s : state) =>
            exists itp : Ltal -> assert,
              DSpec.lookup D rho_tal <@ itp @> 
              /\ itp (mk_Ltal (tal0.RfTy.upd Y rd tal0.valty_int)) psi0 s) 
           / OPsi >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  unfold lift.
  split_l.
  destruct Hwfs as [itp [Hin' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique Hin Hin').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear Hin' Hitpeq.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  exists itptal.
  split_l.
  simpl.
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  eapply tal0.regfile_update; eauto; 
    apply tal0.valok_int.

  (* case wfis_addu_mov *)
  apply wfis_addu.
  unfold lift, itptal, BigInterp.
  exists <<
           (fun (psi0 : chspec) (s : state) =>
            exists itp : Ltal -> assert,
              DSpec.lookup D rho_tal <@ itp @> 
              /\ itp (mk_Ltal (tal0.RfTy.upd Y rd t)) psi0 s) 
           / OPsi >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  unfold lift.
  split_l.
  destruct Hwfs as [itp [Hin' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique Hin Hin').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear Hin' Hitpeq.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  exists itptal.
  split_l.
  simpl.
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  breg_simpl; regfile.simplgs;  simpl.
  inversion_clear Hrt.
  eapply tal0.regfile_update; eauto.
  
  (* case wfis_addiu *)
  apply wfis_addiu.
  unfold lift, itptal, BigInterp.
  exists <<
    (fun (psi0 : chspec) (s : state) =>
      exists itp : Ltal -> assert,
        DSpec.lookup D rho_tal <@ itp @> 
        /\ itp (mk_Ltal (tal0.RfTy.upd Y rd tal0.valty_int)) psi0 s) 
    / OPsi >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  unfold lift.
  split_l.
  destruct Hwfs as [itp [Hin' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique Hin Hin').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear Hin' Hitpeq.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]]. 
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  exists itptal.
  split_l.
  simpl.
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  breg_simpl; regfile.simplgs;  simpl.
  eapply tal0.regfile_update; eauto.
  apply tal0.valok_int; auto.

  (* case wfis_addiu_movi *)
  apply wfis_addiu.
  unfold lift, itptal, BigInterp.
  exists 
    <<
    (fun (psi0 : chspec) (s : state) =>
      exists itp : Ltal -> assert,
        DSpec.lookup D rho_tal <@ itp @> 
        /\ itp (mk_Ltal (tal0.RfTy.upd Y rd 
          (tal0.valty_code Gamma''))) psi0 s)
    / OPsi 
    >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  unfold lift.
  split_l.
  destruct Hwfs as [itp [Hin' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique Hin Hin').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear Hin' Hitpeq.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  exists itptal; split_l; simpl.
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  breg_simpl; regfile.simplgs;  simpl.
  eapply tal0.regfile_update; eauto.
  apply tal0.valok_lab; auto.
  eapply sink_sub_2; eauto.

  (* case wfis_subu *)
  apply wfis_subu.
  unfold lift, itptal, BigInterp.
  exists 
    <<
    (fun (psi0 : chspec) (s : state) =>
      exists itp : Ltal -> assert,
        DSpec.lookup D rho_tal <@ itp @> 
        /\ itp (mk_Ltal (tal0.RfTy.upd Y rd 
          tal0.valty_int)) psi0 s)
    / OPsi 
    >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  unfold lift.
  split_l.
  destruct Hwfs as [itp [Hin' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique Hin Hin').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear Hin' Hitpeq.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  exists itptal; split_l; simpl.
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  eapply tal0.regfile_update; eauto; 
    apply tal0.valok_int.

  (* case wfis_subiu *)
  apply wfis_subiu.
  unfold lift, itptal, BigInterp.
  exists
    <<
    (fun (psi0 : chspec) (s : state) =>
      exists itp : Ltal -> assert,
        DSpec.lookup D rho_tal <@ itp @> 
        /\ itp (mk_Ltal (tal0.RfTy.upd Y rd 
          tal0.valty_int)) psi0 s)
    / OPsi 
    >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  unfold lift.
  split_l.
  destruct Hwfs as [itp [Hin' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique Hin Hin').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear Hin' Hitpeq.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  exists itptal; split_l; simpl.
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  eapply tal0.regfile_update; eauto.
  apply tal0.valok_int; auto.

  (* case wfis_lw *)
  apply wfis_lw.
  unfold lift, itptal, BigInterp.
  exists 
    <<
    (fun (psi0 : chspec) (s : state) =>
      exists itp : Ltal -> assert,
        DSpec.lookup D rho_tal <@ itp @> 
        /\ itp (mk_Ltal (tal0.RfTy.upd Y rt t))
        psi0 s)
    / OPsi 
    >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  destruct Hwfs as [itp [Hin' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique Hin Hin').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear Hin' Hitpeq.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  assert (Hval :=tal0.reg_typing HR Hrs).
  destruct (tal0.heapval_typing HH1 Ht Hval)
    as [v [Hvin Hv]].
  rewrite (Heap.join_lookup_lookup_l H12 Hvin).
  unfold lift.
  split_l.
  exists itptal; split_l; simpl.
  exists H1; exists H2.
  split_l.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  exists Phi.
  split_l.
  eapply tal0.regfile_update; eauto.

  (* case wfis_sw *)
  apply wfis_sw.
  unfold lift, itptal, BigInterp.
  exists 
    <<
    (fun (psi0 : chspec) (s : state) =>
      exists itp : Ltal -> assert,
        DSpec.lookup D rho_tal <@ itp @> 
        /\ itp (mk_Ltal (tal0.RfTy.upd Y rs
          (tal0.valty_tup (tal0.TupTy.upd o i (tal0.init, t)))
        )) psi0 s)
    / OPsi 
    >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl. 
  destruct Hwfs as [itp [Hin' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique Hin Hin').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear Hin' Hitpeq.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
  assert (Hval :=tal0.reg_typing HR Hrs).
  destruct (tal0.heapval_exists HH1 Hct Hval)
    as [v Hvin].
  rewrite (Heap.join_lookup_lookup_l H12 Hvin).
  unfold lift.
  split_l.
  exists itptal; split_l; simpl.
  exists (Heap.set H1 
    (Regfile.get rf rs + i) (Regfile.get rf rt)).
  exists H2.
  split_l.
  eapply Heap.join_set_l; eauto.
  eapply Heap.lookup_indom; eauto.
  split_r using 
    (eapply open_r_only_assert_heap; eauto).
  destruct (tal0.heap_typing_reg_tying_ex_t
    HH1 HR Hrs Hct) as [ct' [Hct'in Hsub']].
  set (addr := (Regfile.get rf rs + i)) in * .
  set (Phi' := tal0.DHSpec.set Phi addr 
    (tal0.init, t)).
  exists psi'.
  split_l.
  exists Phi'.
    assert (Hx := tal0.DHSpec.lookup_canupd 
      (m:=Phi) (a:=addr) 
      (b:= ct') (tal0.init, t) Hct'in).
    assert (Hsubx := tal0.ceTy_sub_init_ct 
      t ct ct' Hsub' Hsub).
    split.
    refine (tal0.heap_update_WF_H HH1 Hsubx Hct'in Hx _).
    apply tal0.ceok_init.
    eapply tal0.reg_typing; eauto.
    eapply tal0.heap_update_WF_R; eauto.
    eapply tal0.heap_update_WF_R_reg; eauto.

  (* case wfis_blt *)
  apply wfis_blt.
  exists 
    <<
    (fun (psi0 : chspec) (s : state) =>
      exists itp : Ltal -> assert,
        DSpec.lookup D rho_tal <@ itp @> 
        /\ itp (mk_Ltal Y) psi0 s)
    / OPsi 
    >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  bnat_blt_dec (Regfile.get rf rs) (Regfile.get rf rt).
  exists <* rho_tal, (mk_Ltal Gamma'') *>.
    split_l.
    apply (OCHSpec.lookup_sub_lookup 
      (Hlift1 f Gamma'' Hf) Hpsub).
    simpl.
    rewrite Hblt.
    exists itptal.
    split_l.
    simpl.
    destruct Hwfs as [itp [Hin' Hwfs]].
    assert (Hitpeq := DSpec.lookup_unique Hin Hin').
    inj_hyp Hitpeq.
    deprewrite <- Hitpeq in Hwfs.
    clear Hin' Hitpeq.
    destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
    destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
    exists H1; exists H2.
    split_l.
    split_r using 
      (eapply open_r_only_assert_heap; eauto).
    exists psi'.
    split_l.
    exists Phi.
    split_l.
    inversion_clear Hsub.
    eapply tal0.regfile_weakening; eauto.
  simpl.
    rewrite Hnblt.
    unfold lift.
    split_l.
    destruct Hwfs as [itp [Hin' Hwfs]].
    assert (Hitpeq := DSpec.lookup_unique Hin Hin').
    inj_hyp Hitpeq.
    deprewrite <- Hitpeq in Hwfs.
    clear Hin' Hitpeq.
    destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
    destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
    exists itptal; split_l; simpl.
    exists H1; exists H2.
    split_l.
    split_r using 
      (eapply open_r_only_assert_heap; eauto).
    exists psi'.
    split_l.
    exists Phi.
    split_lr.

  (* case wfis_bgtz *)
  apply wfis_bgtz.
  exists 
    <<
    (fun (psi0 : chspec) (s : state) =>
      exists itp : Ltal -> assert,
        DSpec.lookup D rho_tal <@ itp @> 
        /\ itp (mk_Ltal Y) psi0 s)
    / OPsi 
    >>.
  split_r using (eapply IHis; eauto).
  intros Psi s [Hpsub Hwfs].
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  bnat_blt_dec O (Regfile.get rf rs).
  exists <* rho_tal, (mk_Ltal Gamma'') *>.
    split_l.
    apply (OCHSpec.lookup_sub_lookup 
      (Hlift1 f Gamma'' Hf) Hpsub).
    unfold nextS.
    subst s.
    cbv iota beta.
    rewrite Hblt.
    simpl.
    exists itptal.
    split_l.
    simpl.
    destruct Hwfs as [itp [Hin' Hwfs]].
    assert (Hitpeq := DSpec.lookup_unique Hin Hin').
    inj_hyp Hitpeq.
    deprewrite <- Hitpeq in Hwfs.
    clear Hin' Hitpeq.
    destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
    destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
    exists H1; exists H2.
    split_l.
    split_r using 
      (eapply open_r_only_assert_heap; eauto).
    exists psi'.
    split_l.
    exists Phi.
    inversion_clear Hsub.
    split_l.
    eapply tal0.regfile_weakening; eauto.
  unfold nextS.
    subst s; cbv iota beta.
    rewrite Hnblt.
    unfold lift.
    split_l.
    destruct Hwfs as [itp [Hin' Hwfs]].
    assert (Hitpeq := DSpec.lookup_unique Hin Hin').
    inj_hyp Hitpeq.
    deprewrite <- Hitpeq in Hwfs.
    clear Hin' Hitpeq.
    destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
    destruct HS as [psi' [Hsink [Phi [HH1 HR]]]].
    exists itptal; split_l; simpl.
    exists H1; exists H2.
    split_l.
    split_r using 
      (eapply open_r_only_assert_heap; eauto).
    exists psi'.
    split_l.
    exists Phi.
    split_lr.
Qed.

Lemma D_tal_describe_lift_psi :
  forall OPsi psi,
    tal_chspec_lift psi OPsi
    -> describe D_tal OPsi.
Proof.
  intros OPsi psi.
  unfold tal_chspec_lift.
  intros H f rho lty theta Hin.
  destruct H as [H1 H2].
  destruct (H2 f <*rho,theta*> Hin) as [Y [Hpi Hin']].
  injection Hpi.
  intros Heq1 Heq2 Heq3.
  subst.
  exists itptal.
  apply DSpec.get_lookup.
  unfold D_tal, DSpec.get, DSpecDef.get.
  bnat_rew (beq_nat rho_tal rho_tal = true).
  trivial.
Qed.

Lemma tal_wfch_ocap_wfch :
  forall ch psi psi' OPsi OPsi',
    tal_chspec_lift psi OPsi
    -> tal_chspec_lift psi' OPsi'
    -> tal_wfch psi psi' ch
    -> ocap_wfch D_tal OPsi OPsi' ch.
Proof.
  intros ch psi psi' OPsi OPsi'.
  intros Hpsi Hpsi' Hch.
  unfold WF_C.
  split.
    split.
    eapply D_tal_describe_lift_psi; eauto.
    eapply D_tal_describe_lift_psi; eauto.
  intros f pi Hin.
  unfold tal_chspec_lift in Hpsi'.
  unfold tal0.WF_C in Hch.
  destruct Hpsi' as [_ Hpsi'].
  destruct (Hpsi' f pi Hin) as [Y [Hpi Hin']].
  subst pi.
  
  destruct (Hch f Y (tal0.CHSpec.lookup_get Hin')) as [is [Hgis Hwfis]].
  exists is; split_l.
  apply tal_wfis_ocap_wfis_2 with psi.
  apply DSpec.get_lookup.
  unfold D_tal, DSpec.get, DSpecDef.get.
  bnat_rew (beq_nat rho_tal rho_tal = true).
  trivial.
  trivial.
  trivial.
Qed.

Theorem tal_sound :
  forall ch psi psi' OPsi OPsi',
    tal_chspec_lift psi OPsi
    -> tal_chspec_lift psi' OPsi'
    -> tal_wfch psi psi' ch
    -> ocap_wfch D_tal OPsi OPsi' ch.
Proof tal_wfch_ocap_wfch.

End TalOcap.