(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* newpair                                                     *)
(*     in SCAP world                                           *)
(*                                                             *)
(***************************************************************)

(* $Id: scap_newpair.v,v 1.7 2006/07/26 07:48:27 ssg05 Exp $ *)

Require Import optionT.
Require Import machine.
Require Import scap.
Require Import cdsa.
Require Import scap_2.
Require Import extac.

Open Local Scope tm_scope.
Open Local Scope scap_scope.
Open Local Scope seplogic_scope.

(* lables *)

Notation NEWPAIR := 0005.

Notation MALLOC := 0010.
Notation MITER := 0015.
Notation MCOMP := 0017.
Notation MRETPTR := 0025.
Notation MSPLIT := 0027.
Notation MNEXT := 0032.
Notation MLPRV := 0035.
Notation MRET := 0037.

(* code *)

(* Notation FLIST := 100. *)
Parameter FLIST : nat.
(* reg alias *)
Notation Local reqsize := a0.
Notation Local actsize := s0.
Notation Local retptr := v0.

Definition newpair_code :=
(* 0005 *)  addiu a0 zero 2 ; 
(* 0006 *)  j MALLOC;;.

Definition malloc_code :=
(* 0010 *)  addiu actsize reqsize HDSIZE;
(* 0011 *)  addiu prev zero NULL;
(* 0012 *)  lw p_ptr FLIST zero;
(* 0013 *)  addiu retptr zero NULL;
(* 0014 *)  j MITER;;.

Definition miter_code :=
(* 0015 *)  blt zero p_ptr MCOMP;
(* 0016 *)  j MITER;;.

Definition mcomp_code :=
(* 0017 *)  lw t0 SZ p_ptr;
(* 0018 *)  addiu t1 actsize HDSIZE;    
(* 0019 *)  blt t1 t0 MSPLIT;
(* 0020 *)  blt t0 actsize MNEXT;
(* 0021 *)  lw t0 NXT p_ptr;
(* 0022 *)  blt zero prev MLPRV;
(* 0023 *)  sw t0 FLIST zero;
(* 0024 *)  j MRETPTR;;.

Definition mretptr_code :=
(* 0025 *)  addiu retptr p_ptr HDSIZE;
(* 0026 *)  j MRET;;.

Definition msplit_code :=
(* 0027 *)  subu t0 t0 actsize;
(* 0028 *)  sw t0 SZ p_ptr;
(* 0029 *)  addu p_ptr p_ptr t0;
(* 0030 *)  sw actsize SZ p_ptr;
(* 0031 *)  j MRETPTR;;.

Definition mnext_code :=
(* 0032 *)  addu prev p_ptr zero;
(* 0033 *)  lw p_ptr NXT p_ptr;
(* 0034 *)  j MITER;;.

Definition mlprv_code :=
(* 0035 *)  sw t0 NXT prev;
(* 0036 *)  j MRETPTR;;.

Definition mret_code :=
(* 0037 *)  jr ra;;.

(* code heap *)

Definition C : codeheap := 
(* NEWPAIR *)
  (0005,   addiu a0 zero 2              ) ::
  (0006,   j MALLOC                     ) ::

(* MALLOC *)
  (0010,   addiu actsize reqsize HDSIZE ) ::
  (0011,   addiu prev zero NULL         ) ::              
  (0012,   lw p_ptr FLIST zero          ) ::
  (0013,   addiu retptr zero NULL       ) ::
  (0014,   j MITER                      ) ::

(* MITER *)
  (0015,   blt zero p_ptr MCOMP         ) ::
  (0016,   j MITER                      ) ::

(* MCOMP *) 
  (0017,   lw t0 SZ p_ptr               ) ::                      
  (0018,   addiu t1 actsize HDSIZE      ) ::
  (0019,   blt t1 t0 MSPLIT             ) ::
  (0020,   blt t0 actsize MNEXT         ) ::
  (0021,   lw t0 NXT p_ptr              ) ::                     
  (0022,   blt zero prev MLPRV          ) ::
  (0023,   sw t0 FLIST zero             ) ::
  (0024,   j MRETPTR                    ) ::

(* MRETPTR *)
  (0025,   addiu retptr p_ptr HDSIZE    ) ::
  (0026,   j MRET                       ) ::

(* MSPLIT *)
  (0027,   subu t0 t0 actsize           ) ::
  (0028,   sw t0 SZ p_ptr               ) ::
  (0029,   addu p_ptr p_ptr t0          ) ::
  (0030,   sw actsize SZ p_ptr          ) ::
  (0031,   j MRETPTR                    ) ::

(* MNEXT *)
  (0032,   addu prev p_ptr zero         ) ::
  (0033,   lw p_ptr NXT p_ptr           ) ::
  (0034,   j MITER                      ) ::

(* MLPRV *)
  (0035,   sw t0 NXT prev               ) ::
  (0036,   j MRETPTR                    ) ::

(* MRET *)
  (0037,   jr ra                        ) ::
  nil.

Lemma C_get_5 : CHeap.get C 5 = Some (addiu a0 zero 2).
Proof. compute; trivial. Qed.

Lemma C_get_6 : CHeap.get C 6 = Some (j MALLOC).
Proof. compute; trivial. Qed.

Lemma C_get_10 : CHeap.get C 10 = Some (addiu actsize reqsize HDSIZE).
Proof. compute; trivial. Qed.

Lemma C_get_11 : CHeap.get C 11 = Some (addiu prev zero NULL).
Proof. compute; trivial. Qed.

Lemma C_get_12 : CHeap.get C 12 = Some (lw p_ptr FLIST zero).
Proof. compute; trivial. Qed.

Lemma C_get_13 : CHeap.get C 13 = Some (addiu retptr zero NULL).
Proof. compute; trivial. Qed.

Lemma C_get_14 : CHeap.get C 14 = Some (j MITER).
Proof. compute; trivial. Qed.

Lemma C_get_15 : CHeap.get C 15 = Some (blt zero p_ptr MCOMP).
Proof. compute; trivial. Qed.

Lemma C_get_16 : CHeap.get C 16 = Some (j MITER).
Proof. compute; trivial. Qed.

Lemma C_get_17 : CHeap.get C 17 = Some (lw t0 SZ p_ptr).
Proof. compute; trivial. Qed.

Lemma C_get_18 : CHeap.get C 18 = Some (addiu t1 actsize HDSIZE).
Proof. compute; trivial. Qed.

Lemma C_get_19 : CHeap.get C 19 = Some (blt t1 t0 MSPLIT).
Proof. compute; trivial. Qed.

Lemma C_get_20 : CHeap.get C 20 = Some (blt t0 actsize MNEXT).
Proof. compute; trivial. Qed.

Lemma C_get_21 : CHeap.get C 21 = Some (lw t0 NXT p_ptr).
Proof. compute; trivial. Qed.

Lemma C_get_22 : CHeap.get C 22 = Some (blt zero prev MLPRV).
Proof. compute; trivial. Qed.

Lemma C_get_23 : CHeap.get C 23 = Some (sw t0 FLIST zero).
Proof. compute; trivial. Qed.

Lemma C_get_24 : CHeap.get C 24 = Some (j MRETPTR).
Proof. compute; trivial. Qed.

Lemma C_get_25 : CHeap.get C 25 = Some (addiu retptr p_ptr HDSIZE).
Proof. compute; trivial. Qed.

Lemma C_get_26 : CHeap.get C 26 = Some (j MRET).
Proof. compute; trivial. Qed.

Lemma C_get_27 : CHeap.get C 27 = Some (subu t0 t0 actsize).
Proof. compute; trivial. Qed.

Lemma C_get_28 : CHeap.get C 28 = Some (sw t0 SZ p_ptr).
Proof. compute; trivial. Qed.

Lemma C_get_29 : CHeap.get C 29 = Some (addu p_ptr p_ptr t0).
Proof. compute; trivial. Qed.

Lemma C_get_30 : CHeap.get C 30 = Some (sw actsize SZ p_ptr).
Proof. compute; trivial. Qed.

Lemma C_get_31 : CHeap.get C 31 = Some (j MRETPTR).
Proof. compute; trivial. Qed.

Lemma C_get_32 : CHeap.get C 32 = Some (addu prev p_ptr zero).
Proof. compute; trivial. Qed.

Lemma C_get_33 : CHeap.get C 33 = Some (lw p_ptr NXT p_ptr).
Proof. compute; trivial. Qed.

Lemma C_get_34 : CHeap.get C 34 = Some (j MITER).
Proof. compute; trivial. Qed.

Lemma C_get_35 : CHeap.get C 35 = Some (sw t0 NXT prev).
Proof. compute; trivial. Qed.

Lemma C_get_36 : CHeap.get C 36 = Some (j MRETPTR).
Proof. compute; trivial. Qed.

Lemma C_get_37 : CHeap.get C 37 = Some (jr ra).
Proof. compute; trivial. Qed.

Lemma C_get_100 : CHeap.get C 100 = None.
Proof. compute; trivial. Qed.

Lemma C_get_101 : CHeap.get C 101 = None.
Proof. compute; trivial. Qed.

Lemma C_get_102 : CHeap.get C 102 = None.
Proof. compute; trivial. Qed.

Lemma C_getis_newpair_code : getis C NEWPAIR newpair_code.
Proof. getis_solve. Qed.

Lemma C_getis_malloc_code : getis C MALLOC malloc_code.
Proof. getis_solve. Qed.

Lemma C_getis_miter_code : getis C MITER miter_code.
Proof. getis_solve. Qed.

Lemma C_getis_mcomp_code : getis C MCOMP mcomp_code.
Proof. getis_solve. Qed.

Lemma C_getis_mretptr_code : getis C MRETPTR mretptr_code.
Proof. getis_solve. Qed.

Lemma C_getis_msplit_code : getis C MSPLIT msplit_code.
Proof. getis_solve. Qed.

Lemma C_getis_mnext_code : getis C MNEXT mnext_code.
Proof. getis_solve. Qed.

Lemma C_getis_mlprv_code : getis C MLPRV mlprv_code.
Proof. getis_solve. Qed.

Lemma C_getis_mret_code : getis C MRET mret_code.
Proof. getis_solve. Qed.

(* specs *)

(* Definition FList := *)
(*   lexists n, flist fl n. *)

Definition newpair_pre s :=
  exists fv, 
    s |||- FLIST |-> fv ** Good fv ** top.

Definition newpair_guar s s' :=
  rget s ra = S_pc s' /\
  forall A,
    (exists fv,
      s |||- FLIST |-> fv ** Good fv ** A)
    -> (exists fv, exists lnk, exists siz,
      siz >= (2 + 2) /\
      s' |||- MBlk (rget s' retptr) lnk siz ** FLIST |-> fv ** 
        Good fv ** A).
  
Definition newpair_spec := [newpair_pre & newpair_guar].

Definition malloc_pre s :=
  rget s a0 > 0 
  /\ exists fv, 
    s |||- FLIST |-> fv ** Good fv ** top.

Definition malloc_guar s s' :=
  rget s ra = S_pc s' /\
  forall A,
    (exists fv,
      s |||- FLIST |-> fv ** Good fv ** A)
    -> (exists fv, exists lnk, exists siz,
      siz >= (rget s a0 + 2) /\
      s' |||- MBlk (rget s' retptr) lnk siz ** 
        FLIST |-> fv ** Good fv ** A).

Definition miter_pre s :=
  rget s a0 > 0 
  /\ rget s actsize = rget s a0 + 2 
  /\ rget s retptr = NULL 
  /\ exists fv, 
    s |||- FLIST |-> fv ** MidL fv (rget s prev) (rget s p_ptr) ** top.

Definition miter_guar s s' :=
  rget s ra = S_pc s' 
  /\ forall A,
    (exists fv,  
      s |||- FLIST |-> fv ** MidL fv (rget s prev) (rget s p_ptr) ** A) 
    -> 
    (exists fv, exists lnk, exists siz,
      siz >= (rget s a0 + 2) /\
      s' |||- MBlk (rget s' retptr) lnk siz ** 
        FLIST |-> fv ** Good fv ** A).

Definition mcomp_pre s :=
  rget s a0 > 0 
  /\ rget s actsize = rget s a0 + 2 
  /\ rget s retptr = NULL 
  /\ rget s p_ptr > NULL 
  /\ exists fv,  
    s |||- FLIST |-> fv ** MidL fv (rget s prev) (rget s p_ptr) ** top.

Notation mcomp_guar := miter_guar.

Definition msplit_pre s :=
  rget s a0 > 0 
  /\ rget s actsize = rget s a0 + 2 
  /\ rget s p_ptr > NULL 
  /\ rget s t0 > rget s actsize + 2 
  /\ exists fv,  
    exists q, exists n,
      (rget s p_ptr < q \/ q = NULL) /\
      s |||- FLIST |-> fv ** EndL fv (rget s prev) (rget s p_ptr) ** 
        MBlk (rget s p_ptr + 2) q (rget s t0) ** MBlkLst n q NULL ** top.

Definition msplit_guar s s' :=
  rget s ra = S_pc s'/\ 
  forall A,
    (exists fv,  
      exists q, exists n,
        (rget s p_ptr < q \/ q = NULL) /\
        s |||- FLIST |-> fv ** EndL fv (rget s prev) (rget s p_ptr) ** 
	  MBlk (rget s p_ptr + 2) q (rget s t0) ** MBlkLst n q NULL ** A) 
    ->
    (exists fv, exists lnk, exists siz,
      siz >= (rget s a0 + 2) /\
      s' |||- MBlk (rget s' retptr) lnk siz ** 
        FLIST |-> fv ** Good fv ** A).

Definition mnext_pre s :=
  rget s a0 > 0 
  /\ rget s actsize = rget s a0 + 2 
  /\ rget s p_ptr > NULL 
  /\ rget s retptr = NULL 
  /\ exists fv,  
    exists q, exists n,
      (rget s p_ptr < q \/ q = NULL) /\
      s |||- FLIST |-> fv ** EndL fv (rget s prev) (rget s p_ptr) ** 
        MBlk (rget s p_ptr + 2) q (rget s t0) ** MBlkLst n q NULL ** top.

Notation mnext_guar := msplit_guar.

Definition mlprv_pre s :=
  rget s a0 > 0 
  /\ rget s actsize = rget s a0 + 2 
  /\ rget s p_ptr > NULL 
  /\ rget s prev > NULL 
  /\ exists fv,  
    exists n, exists siz,
      (rget s p_ptr < rget s t0 \/ rget s t0 = NULL) /\ 
      (siz >= rget s actsize) /\
      s |||- FLIST |-> fv ** EndL fv (rget s prev) (rget s p_ptr) ** 
        MBlk (rget s p_ptr + 2) (rget s t0) siz ** 
        MBlkLst n (rget s t0) NULL ** top.

Definition mlprv_guar s s' :=
  rget s ra = S_pc s'/\ 
  forall A,
    (exists fv, 
      exists n, exists siz,
      (rget s p_ptr < rget s t0 \/ rget s t0 = NULL) /\ 
      (siz >= rget s actsize) /\
      s |||- FLIST |-> fv ** 
        EndL fv (rget s prev) (rget s p_ptr) ** 
	MBlk (rget s p_ptr + 2) (rget s t0) siz ** 
        MBlkLst n (rget s t0) NULL ** A) 
    -> 
    (exists fv, exists lnk, exists siz,
    siz >= (rget s a0 + 2) /\
    s' |||- MBlk (rget s' retptr) lnk siz ** 
      FLIST |-> fv ** Good fv ** A).

Definition mretptr_pre s :=
  rget s a0 > 0 
  /\ rget s actsize = rget s a0 + 2
  /\ exists fv, exists siz, exists lnk,
    (siz >= rget s actsize) /\
    s |||- FLIST |-> fv ** 
      MBlk (rget s p_ptr + 2) lnk siz ** Good fv ** top.

Definition mretptr_guar s s' :=
  rget s ra = S_pc s' /\ 
  forall A,
    (exists fv, exists siz, exists lnk,
      (siz >= rget s actsize) /\
      s |||- FLIST |-> fv ** 
        MBlk (rget s p_ptr + 2) lnk siz ** 
        Good fv ** A) 
    ->
    (exists fv, exists lnk, exists siz,
      siz >= (rget s a0 + 2) /\
      s' |||- MBlk (rget s' retptr) lnk siz ** 
        FLIST |-> fv ** Good fv ** A).

Definition mret_pre (s : state) := True.

Definition mret_guar (s s' : state) := 
  let (h, rf, pc) := s in <.h, rf, rget s ra.> = s'.

(* code heap spec *)
Definition psi : chspec := fun l =>
  if (beq_nat l NEWPAIR) then SomeT (mk_cdspec newpair_pre newpair_guar) else
  if (beq_nat l MALLOC) then SomeT (mk_cdspec malloc_pre malloc_guar) else
  if (beq_nat l MITER) then SomeT (mk_cdspec miter_pre miter_guar) else
  if (beq_nat l MNEXT) then SomeT (mk_cdspec mnext_pre mnext_guar) else
  if (beq_nat l MRETPTR) then SomeT (mk_cdspec mretptr_pre mretptr_guar) else
  if (beq_nat l MLPRV) then SomeT (mk_cdspec mlprv_pre mlprv_guar) else
  if (beq_nat l MSPLIT) then SomeT (mk_cdspec msplit_pre msplit_guar) else
  if (beq_nat l MCOMP) then SomeT (mk_cdspec mcomp_pre mcomp_guar) else
  if (beq_nat l MRET) then SomeT (mk_cdspec mret_pre mret_guar) else
    NoneT.

(* proof *)

Ltac unfoldpred :=
  unfold newpair_pre, newpair_guar;
  unfold malloc_pre, malloc_guar; 
  unfold miter_pre, miter_guar; 
  unfold mnext_pre, mretptr_pre; 
  unfold msplit_pre, msplit_guar; 
  unfold mlprv_pre, mlprv_guar; 
  unfold mcomp_pre, mretptr_guar; 
  unfold mret_pre, mret_guar.

(* Axiom newpair_ok : wfis psi newpair_pre newpair_guar newpair_code. *)
(* Axiom malloc_ok : wfis psi malloc_pre malloc_guar malloc_code. *)
(* Axiom miter_ok : wfis psi miter_pre miter_guar miter_code. *)
(* Axiom mcomp_ok : wfis psi mcomp_pre mcomp_guar mcomp_code. *)
(* Axiom mretptr_ok : wfis psi mretptr_pre mretptr_guar mretptr_code. *)
(* Axiom msplit_ok : wfis psi msplit_pre msplit_guar msplit_code. *)
(* Axiom mnext_ok : wfis psi mnext_pre mnext_guar mnext_code. *)
(* Axiom mlprv_ok : wfis psi mlprv_pre mlprv_guar mlprv_code. *)
(* Axiom mret_ok : wfis psi mret_pre mret_guar mret_code. *)

Theorem newpair_ok : wfis psi newpair_pre newpair_guar newpair_code.
Proof.
  simplvc.
  unfoldpred.
  intros s. 
  destruct s as [hx rfx pcx].
  intros.
  split.
    split_r.
    simplst.
    auto with arith.
  intros [hx' rfx' pcx'].
  unfold newpair_guar, rget.
  simplst.
  intros [Hra H1].
  split_l.
  intros.
  substH H1 with (H1 A H0).
  destruct H1 as [fv [lnk [siz [Hsiz Hblk]]]].
  exists fv; exists lnk; exists siz.
  split_lr.
Qed.
  
Theorem malloc_ok : wfis psi malloc_pre malloc_guar malloc_code.
Proof.
  simplvc.

  unfoldpred.
  intros s. 
  destruct s as [hx rfx pcx].
  simplst.
  intros [R [fv Hp]].
  simpllw fv 1.
  simplst.
  intuition.
  exists fv.
  simplsep 3 3.
  apply (Good_MidL Hp).

  apply H1.
  destruct H.
  hvinject 1 1; subst.
  clear Hp; exists x; simplsep 3 3. 
  apply (Good_MidL H).
Qed.


Lemma miter_ok : wfis psi miter_pre miter_guar miter_code.
Proof.
  simplvc.
  
  unfoldpred.
  intros [hx rfx pcx].
  unfold rget, rset.
  simplst.
  intros [R1 [R2 [R3 [fv Hp]]]].
  destruct (eq_dec (Regfile.get rfx p_ptr) O)
    as [Heq | Hneq].
    rewrite Heq.
    rewrite Heq in Hp.
    bnat_rew (blt_nat O O = false).
    split.
      split_l; split_l; split_l; exists fv; trivial.
    intros [hx' rfx' pcx'].
    intros H.
    intuition.
(*   reflexivity. *)
(*   destruct H0. *)
(*   right; exists x; split; auto. *)
(*   simplsep 3 3. *)
(*   apply (MidL_Good H). *)
  bnat_rew (blt_nat O (Regfile.get rfx p_ptr) = true).
  assert (H' : exists n, Regfile.get rfx p_ptr = S n).
    destruct (Regfile.get rfx p_ptr).
    tauto.
    exists n; trivial.
  destruct H' as [n0 H'].
  rewrite H'.
  intuition.
  rewrite H' in Hp.
  exists fv; apply Hp.
Qed.

Lemma mcomp_ok : wfis psi mcomp_pre mcomp_guar mcomp_code.
Proof.
  simplvc.
  intros [h rf pc] [R1 [R2 [R3 [R4 [fv Hp]]]]].
  simplst.
  unfold MidL in Hp; sepelim Hp.
  destruct x as [ | n]; simpl in Hp; sepelim Hp.
    elimtype False; omega.
  unfold MBlk in Hp; sepelim Hp; simplpm2h Hp.
  simpllw x0 4.
  regfile.simplgs.
  bnat_blt_dec  (Regfile.get rf actsize + 2) x0.
    repeat (split; trivial).
          bnat_elim.
          auto.
        exists fv; exists x; exists n.
        split; auto.
        unfold MBlk; simplpm2; simplsep 5 5.
      unfold msplit_guar in H.
      tauto.
    destruct H as [_ G].
    intros P [fv' Hp'].
    apply G; clear G.
    unfold MidL in Hp'; sepelim Hp'.
    destruct x1 as [ | n1]; simpl in Hp'; sepelim Hp'.
      elimtype False; omega.
    unfold MBlk in Hp'; sepelim Hp'; simplpm2h Hp'.
    hvinject 7 7. subst; clear Hp.
    exists fv'; exists x1; exists n1.
    split; trivial.
    unfold MBlk; simplpm2. 
    abstract simplsep 5 5.

  simplst.
  bnat_blt_dec x0 (Regfile.get rf actsize).
    simplst.
    repeat (split; trivial).
        exists fv; exists x; exists n.
        split; trivial.
        unfold MBlk; simplpm2; simplsep 5 5.
      unfoldgs.
      destruct s' as [h' rf' pc'].
      unfold msplit_guar in H.
      tauto.
    destruct H as [_ G].
    intros P [fv' Hp'].
    apply G; clear G.
    unfold MidL in Hp'; sepelim Hp'.
    destruct x1 as [ | n1]; simpl in Hp'; sepelim Hp'.
      elimtype False; omega.
    unfold MBlk in Hp'; sepelim Hp'; simplpm2h Hp'.
    hvinject 7 7; subst; clear Hp.
    exists fv'; exists x1; exists n1.
    split; trivial.
    unfold MBlk; simplpm2; abstract simplsep 5 5.
  
  simplpm.
  simpllw x 3.
  destruct (eq_dec (Regfile.get rf prev) O) 
    as [Heq | Hneq].
    regfile.simplgs.
    bnat_rew (blt_nat 0 (Regfile.get rf prev) = false).
    simplpm.
    unfold EndL in Hp. 
    rewrite Heq in Hp.
    bnat_rew (beq_nat O O = true) in Hp.
    simpl in Hp. 
    sepelim Hp.
    subst fv.
    simpllw (Regfile.get rf p_ptr) 1.
    repeat (split; trivial).
        exists x; exists x0; exists x.
        split; trivial.
        unfold rget.
        regfile.simplgs.
        bnat_elim.
        trivial.
        unfold Good; lexintro 4; exists n.
        unfold MBlk; simplpm2.
        simpl; abstract simplsw 6 6 (Regfile.get rf p_ptr).
      unfoldgs.
      unfold mretptr_guar in H.
      tauto.
    destruct H as [_ G].
    intros P [fv' Hp'].
    apply G; clear G.
    unfold MidL in Hp'; sepelim Hp'.
    unfold EndL in Hp'; simpl in Hp'; sepelim Hp'.
    rewrite Heq in Hp'.
    bnat_rew (beq_nat O O = true) in Hp'.
    sepelim Hp'.
    subst fv'.
    destruct x1 as [ | n1]; simpl in Hp'; sepelim Hp'.
      elimtype False; omega.
    unfold MBlk in Hp'; sepelim Hp'; simplpm2h Hp'.
    hvinject 7 7. subst x x0; clear Hp.
    exists x1; exists x2; exists x1.
    split; trivial.
      unfold rget.
      regfile.simplgs.
      bnat_elim.
      trivial.
    unfold rget.
    regfile.simplgs.
    unfold Good; lexintro 4; exists n1.
    unfold MBlk; simplpm2.

  simpl; abstract simplsw 6 6 (Regfile.get rf p_ptr).
  regfile.simplgs.
  bnat_rew (blt_nat 0 (Regfile.get rf prev) = true).
  assert (H' : exists n, Regfile.get rf prev = S n).
    destruct (Regfile.get rf prev).
      tauto.
    exists n0; trivial.
  destruct H' as [n0 H'].
  unfold miter_guar.
  unfoldgs; regfile.simplgs.
  rewrite H'.
  repeat (split; trivial).
         bnat_elim.
         bnat_elim.
         unfoldgs; regfile.simplgs.
         lwomega.
      exists fv; exists n; exists x0.
      repeat (split; trivial).
        bnat_elim.
        bnat_elim.
        unfoldgs; regfile.simplgs.
        omega.
      unfold MBlk; simplpm2; simplsep 5 5.
    unfold mlprv_guar in H.
    tauto.
  destruct H as [_ G].
  intros P [fv' Hp'].
  apply G; clear G.
  unfold MidL in Hp'; sepelim Hp'.
  destruct x1 as [ | n1]; simpl in Hp'; sepelim Hp'.
    elimtype False; omega.
  unfold MBlk in Hp'; sepelim Hp'; simplpm2h Hp'.
  hvinject 7 7. subst; clear Hp.
  exists fv'; exists n1; exists x0.
  repeat (split; trivial).
    bnat_elim; bnat_elim.
    unfoldgs; regfile.simplgs.
    omega.
  unfoldgs.
  regfile.simplgs.
  rewrite H'.
  unfold MBlk; simplpm2.
  simplsep 5 5.
Qed.

Lemma mretptr_ok : wfis psi mretptr_pre mretptr_guar mretptr_code.
Proof. 
  simplvc.
  intros [h rf pc].
  unfold mretptr_pre, mretptr_guar, mret_pre, mret_guar.
  simplst.
  intuition.
  destruct s' as [h' rf' pc'].
  simplst.
  injection H1.
  intros; trivial.
  clear H2; destruct H3 as [fv [siz [lnk [R Hp]]]].
  exists fv; exists lnk; exists siz.
  split.
  omega.
  destruct s' as [h' rf' pc'].
  injection H1.
  intros.
  subst.
  simplst.
  simplsep 4 4.
Qed.

Lemma msplit_ok : wfis psi msplit_pre msplit_guar msplit_code.
Proof.
  simplvc.
  intros [h rf pc] [R1 [R2 [R3 [R4 [fv [q [n [R5 Hp]]]]]]]].
  simplst.
  unfold MBlk in Hp; sepelim Hp; simplpm2h Hp.
  simpllw (Regfile.get rf t0) 3.
    assert ((S (Regfile.get rf p_ptr)) <> (S (Regfile.get rf p_ptr 
      + (Regfile.get rf t0 - Regfile.get rf actsize)))).
    omega.
    clear H.
    sconjlifth 4 Hp.
    assert (Regfile.get rf t0 - 2 > Regfile.get rf reqsize + 2).
    lwomega.
    hypreplace Hp (buffer_split Hp H).
    destruct Hp as [y [z Hp]].
    rewrite <- R2 in Hp.
    replace (Regfile.get rf p_ptr + 2 
      + (Regfile.get rf t0 - 2 - Regfile.get rf actsize)) 
    with
    (Regfile.get rf p_ptr + (Regfile.get rf t0 - Regfile.get rf actsize)) in Hp;
      [ simpllw z 3| lwomega].
    repeat (split; trivial).
        exists fv; exists (Regfile.get rf actsize); exists y.
        split; auto.
        unfold MBlk; simplpm2.
        simplst.
        simplsw' 4 z.
        sconjlifth 9 Hp.
        hypreplace Hp (cont_update1 Hp (Regfile.get rf t0 
          - Regfile.get rf actsize)).
        replace (Regfile.get rf p_ptr + 2 + (Regfile.get rf t0 - 2 
          - Regfile.get rf reqsize)) with
        (Regfile.get rf p_ptr + (Regfile.get rf t0 - Regfile.get rf actsize) + 2) in Hp;
        [replace (Regfile.get rf reqsize) with (Regfile.get rf actsize - 2) in Hp;
          [simplsep 9 9 | lwomega] | lwomega].
          lwomega.
        apply MidL_Good with (Regfile.get rf prev) (Regfile.get rf p_ptr).
        unfold MidL.
        lexintro 2; exists (S n).
        simpl.
        lexintro 2; exists q.
        lexintro 2; exists (Regfile.get rf t0 - Regfile.get rf actsize).
        unfold MBlk; simplpm2.
        replace (Regfile.get rf t0 - Regfile.get rf actsize - 2) with
          (Regfile.get rf t0 - 2 - Regfile.get rf actsize);
          [simplsep  5 5; omega| lwomega].
      unfold mretptr_guar in H0; simplst.
      tauto.
    destruct H0 as [_ G].
    intros P [fl' [q' [n' [R' Hp']]]].
    apply G; clear G.

    unfold MBlk in Hp'; sepelim Hp'; simplpm2h Hp'.
    sconjlifth 4 Hp'.
    hypreplace Hp' (buffer_split Hp' H).
    destruct Hp' as [y' [z' Hp']].
    unfold rget, hget in Hp'.
    rewrite <- R2 in Hp'.
    replace (Regfile.get rf p_ptr + 2 + (Regfile.get rf t0 - 2 
      - Regfile.get rf actsize)) with
      (Regfile.get rf p_ptr + (Regfile.get rf t0 - Regfile.get rf actsize)) in Hp';
      [idtac | lwomega].
    hvinject 8 8. subst; clear Hp.
    exists fv; exists (Regfile.get rf actsize); exists y'.
    split; auto.
    unfold MBlk; simplpm2.
    unfold hset; simpl.
(*   sconjlift 3. *)
(*   apply cont_update1 with (v1 := z). *)
    simplst.
    simplsw' 6 z.
    sconjlifth 9 Hp'.
    hypreplace Hp' (cont_update1 Hp' (Regfile.get rf t0 - Regfile.get rf actsize)).
    replace (Regfile.get rf p_ptr + 2 + (Regfile.get rf t0 - 2 
      - Regfile.get rf reqsize)) with
      (Regfile.get rf p_ptr + (Regfile.get rf t0 - Regfile.get rf actsize) + 2) in Hp';
      [replace (Regfile.get rf reqsize) with (Regfile.get rf actsize - 2) in Hp';
        [simplsep 9 9 | lwomega] | lwomega].
      lwomega.
    apply MidL_Good with (Regfile.get rf prev) (Regfile.get rf p_ptr).
    unfold MidL.
    lexintro 2; exists (S n').
    simpl.
    lexintro 2; exists q.
    lexintro 2; exists (Regfile.get rf t0 - Regfile.get rf actsize).
    unfold MBlk; simplpm2.
    replace (Regfile.get rf t0 - Regfile.get rf actsize - 2) with
      (Regfile.get rf t0 - 2 - Regfile.get rf actsize);
      [simplsep  5 5; omega| lwomega].
  lwomega.
Qed.

Lemma mnext_ok : wfis psi mnext_pre mnext_guar mnext_code.
Proof.
  simplvc.
  intros [h rf pc] 
    [R1 [R2 [R3 [R4 [fv [q [n [R5 Hp]]]]]]]].
  simplst.
  assert (Hp' := Hp).
  unfold MBlk in Hp; simplpm2h Hp; sepelim Hp.
  simpllw q 4.
  repeat (split; trivial).
      exists fv.
      clear Hp; unfold MidL.
      lexintro 3; exists n.
      simplsep 3 4.
      simplst.
      eapply EndL_add; auto.
      apply Hp'.
    unfold miter_guar in H; simplst.  
    tauto.
  destruct H as [_ G].
  intros P [fv' [q' [n' [R' Hp1]]]].
  apply G; clear G.
  generalize Hp1.
  unfold MBlk in Hp1; simplpm2h Hp1; sepelim Hp1.
  clear Hp'; hvinject 6 6; subst.
  clear Hp1 Hp.
  exists fv'; unfold MidL.
  lexintro 3; exists n'.
  simplsep 3 4.
  eapply EndL_add; auto.
  apply Hp0.
Qed.

Lemma mlprv_ok : wfis psi mlprv_pre mlprv_guar mlprv_code.
Proof.
  simplvc.
  simplst.
  intros [h rf pc] 
    [R1 [R2 [R3 [R4 [fv [n [siz [R5 [R6 Hp]]]]]]]]].
  simplst.
  unfold EndL in Hp.
  rewrite neq_beq_false in Hp; [unfold MBlk in Hp; sepelim Hp | omega].
  simplpm2h Hp.
  simpllw (Regfile.get rf p_ptr) 3.
  repeat (split; trivial).
      exists fv; exists siz; exists (Regfile.get rf t0).
      split; trivial.
      sconjlifth 2 Hp.
      hypreplace Hp (cont_update1 Hp (Regfile.get rf t0)).
      unfold hset; simpl.
      unfold MBlk; simplpm2; simplsep 6 7.
      assert (h0 ||- MBlkLst x fv (Regfile.get rf prev) ** 
        MBlk (Regfile.get rf prev + 2) (Regfile.get rf t0) x0 ** MBlkLst n (Regfile.get rf t0) 0).
        unfold MBlk; simplpm2; simplsep 5 5.
      clear Hp.
      hypreplace H (MBlkLst_add H).
      assert (Regfile.get rf prev < Regfile.get rf t0 \/ Regfile.get rf t0 = 0).
        lwomega.
      substH H with (unit_intro_right (H H0)).
      assocl H.
      substH H with (MBlkLst_append H).
      unfold Good.
      lexintro 1; exists (S x + n).
      apply (unit_elim_right H).
    unfold mretptr_guar in H; simplst.
    tauto.
  destruct H as [_ G].
  intros P [fv' [n' [siz' [R1' [R2' Hp1]]]]].
  apply G; clear G.
  unfold EndL in Hp1.
  simplst.
  rewrite neq_beq_false in Hp1; [unfold MBlk in Hp1; sepelim Hp1 | omega].
  simplpm2h Hp1.
  hvinject 8 8; subst; clear Hp.
  clear R1' R2' A2.
  exists fv'; exists siz; exists (Regfile.get rf t0).
  split; trivial.
  sconjlifth 2 Hp1.
  substH Hp1 with (cont_update1 Hp1 (Regfile.get rf t0)).
  unfold hset; simpl.
  unfold MBlk; simplpm2; simplsep 6 7.
  assert (h0 ||- MBlkLst x1 fv' (Regfile.get rf prev) ** 
    MBlk (Regfile.get rf prev + 2) (Regfile.get rf t0) x2 ** MBlkLst n' (Regfile.get rf t0) 0).
    unfold MBlk; simplpm2; simplsep 5 5.
  clear Hp1.
  substH H with (MBlkLst_add H).
  assert (Regfile.get rf prev < Regfile.get rf t0 \/ Regfile.get rf t0 = 0).
    lwomega.
  substH H with (unit_intro_right (H H0)).
  assocl H.
  substH H with (MBlkLst_append H).
  unfold Good.
  lexintro 1; exists (S x1 + n').
  apply (unit_elim_right H).
Qed.

Lemma mret_ok : wfis psi mret_pre mret_guar mret_code.
Proof.
  simplvc.
  intros [h rf pc].
  unfold mret_pre, mret_guar.
  simplst.
  intuition.
Qed.

Theorem C_ok : wfch psi psi C.
Proof.
  unfold wfch. intros f p g H.

  psi_lookup (NEWPAIR).
  exists newpair_code.
  split_l with C_getis_newpair_code.
  apply newpair_ok.

  psi_lookup (MALLOC).
  exists malloc_code.
  split_l with C_getis_malloc_code.
  apply malloc_ok.

  psi_lookup (MITER).
  exists miter_code.
  split_l with C_getis_miter_code.
  apply miter_ok.

  psi_lookup (MNEXT).
  exists mnext_code.
  split_l with C_getis_mnext_code.
  apply mnext_ok.

  psi_lookup (MRETPTR).
  exists mretptr_code.
  split_l with C_getis_mretptr_code.
  apply mretptr_ok.

  psi_lookup (MLPRV).
  exists mlprv_code.
  split_l with C_getis_mlprv_code.
  apply mlprv_ok.

  psi_lookup (MSPLIT).
  exists msplit_code.
  split_l with C_getis_msplit_code.
  apply msplit_ok.

  psi_lookup (MCOMP).
  exists mcomp_code.
  split_l with C_getis_mcomp_code.
  apply mcomp_ok.

  psi_lookup (MRET).
  exists mret_code.
  split_l with C_getis_mret_code.
  apply mret_ok.

  psi_lookup_not.
Qed.
