(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* MAP Interface                                               *)
(*                                                             *)
(***************************************************************)

(* $Id: MAPINC.v,v 1.4 2006/07/15 02:56:42 ssg05 Exp $ *)



(*   Parameter A : Set. *)
Definition A := MapDef.A.

(*   Axiom eq_dec : forall a a' : A, {a=a'}+{a<>a'}. *)
Definition eq_dec := MapDef.eq_dec.

(*   Parameter B : Set. *)
Definition B := MapDef.B.

(***********************)
(** map                *)
(***********************)

(*   Parameter map : Type. *)
Definition map : B_U := MapDef.map.

(*   Parameter emp_map : map. *)
Definition emp_map : map := MapDef.emp_map.

(*   Parameter sig_map : A -> B -> map. *)
Definition sig_map : A -> B -> map 
  := MapDef.sig_map.
  
(***********************)
(** set                *)
(** get                *)
(***********************)

(*   Parameter get : map -> A -> optionX B. *)
Definition get : map -> A -> optionX B
  := MapDef.get.

Definition get_dec : forall (m : map) (a : A),
  {exists b, get m a = SomeX b} + {get m a = NoneX}
  := MapDef.get_dec.

Definition set : map -> A -> B -> map
  := MapDef.set.

Lemma set_sem : forall m a a' b, 
  match (beq a a') with
    | true => get (set m a b) a = SomeX b
    | false => get (set m a b) a' = get m a'
  end.
Proof MapDef.set_sem.
  
(*   Axiom emp_sem : forall a : A, get emp_map a = NoneX. *)
Lemma emp_sem : forall a : A, get emp_map a = NoneX.
Proof MapDef.emp_sem.

Lemma sig_sem : forall (a a' : A) (b : B),
  get (sig_map a' b) a =
  match (beq a a') with
    | true => SomeX b
    | false => NoneX
  end.
Proof MapDef.sig_sem.


(***********************************************************)
(**                                                        *)
(**    Lemmas                                              *)
(**                                                        *)
(**    Tactics                                             *)
(**                                                        *)
(***********************************************************)

(***********************)
(** indom              *)
(** lookup             *)
(***********************)

Definition indom (m : map) (a : A) := 
  exists b : B, get m a = SomeX b.

Definition lookup (m : map) (a : A) (b : B) := 
  get m a = SomeX b.
  
(***********************)
(** sub                *)
(** meq                *)
(** compat             *)
(***********************)

Definition sub (m1 m2 : map) :=
  forall (a : A) (b : B), 
    lookup m1 a b -> lookup m2 a b.
  
Definition meq (m1 m2 : map) :=
  sub m1 m2 /\ sub m2 m1.

Definition compat (m1 m2 : map) : Prop :=
  forall a : A,
    match get m1 a, get m2 a with
      | SomeX b1, SomeX b2 => b1 = b2
      | _, _ => True
    end.

(************************)
(** disj                *)
(************************)

  Definition disj (m1 m2 : map) : Prop :=
    forall a : A, 
      match get m1 a, get m2 a with
        | SomeX b, NoneX => True
        | NoneX, SomeX b => True
        | NoneX, NoneX => True
        | _ , _ => False
      end.

(************************)
(** merge               *)
(************************)

  Definition merge : map -> map -> map
    := MapDef.merge.

  Lemma merge_sem : forall (m1 m2 : map) (a : A),
    get (merge m1 m2) a
    = match get m1 a, get m2 a with
        | SomeX b1, SomeX b2 => SomeX b1
        | SomeX b1, NoneX => SomeX b1
        | NoneX, SomeX b2 => SomeX b2
        | NoneX, NoneX => NoneX
      end.
  Proof MapDef.merge_sem.


(************************)
(** join                *)
(************************)

  Definition join (m1 m2 m: map) :=
    forall a : A,
      match get m1 a, get m2 a, get m a with
        | NoneX, SomeX b1, SomeX b2 => b1 = b2
        | SomeX b1, NoneX, SomeX b2 => b1 = b2
        | NoneX, NoneX, NoneX => True
        | _, _, _ => False
      end.


(***********************************************************)
(**                                                        *)
(**    Lemmas                                              *)
(**                                                        *)
(**    Tactics                                             *)
(**                                                        *)
(***********************************************************)

(* Ltac eqdec_eq :=   *)
(*   match goal with *)
(*     | H: context [eq_dec ?a ?a] |- _  *)
(*       =>  let He := fresh "He" in *)
(*           let Hne := fresh "Hne" in *)
(*             (destruct (eq_dec a a) as [He | Hne ];  *)
(*               [ clear He | discri ]) *)
(*     | |- context [eq_dec ?a ?a] *)
(*       =>  let He := fresh "He" in *)
(*           let Hne := fresh "Hne" in *)
(*             (destruct (eq_dec a a) as [He | Hne ];  *)
(*               [ clear He | discri ]) *)
(*     | _ => fail 1 "no eq_dec term found " *)
(*   end. *)

(* Ltac eqdec_neq :=   *)
(*   match goal with *)
(*     | H: context [eq_dec ?a ?b] |- _  *)
(*       =>  let He := fresh "He" in *)
(*           let Hne := fresh "Hne" in *)
(*             (destruct (eq_dec a b) as [He | Hne]; *)
(*               [ *)
(*                 match goal with  *)
(*                   | Hneq : not (eq a b) |- _ => *)
(*                     destruct (Hneq He) *)
(*                   | Hneq : not (eq b a) |- _ => *)
(*                     destruct (Hneq (sym_eq He)) *)
(*                   | _ => fail 1 "xx" *)
(*                 end *)
(*               | idtac ]) *)
(*     | |- context [eq_dec ?a ?b] *)
(*       =>  let He := fresh "He" in *)
(*           let Hne := fresh "Hne" in *)
(*             (destruct (eq_dec a b) as [He | Hne]; *)
(*               [ *)
(*                 match goal with  *)
(*                   | Hneq : not (eq a b) |- _ => *)
(*                     destruct (Hneq He) *)
(*                   | Hneq : not (eq b a) |- _ => *)
(*                     destruct (Hneq (sym_eq He)) *)
(*                   | _ => fail *)
(*                 end *)
(*               | idtac]) *)
(*     | _ => fail 1 "no eq_dec term found " *)
(*   end. *)

Lemma beq_refl : forall a, beq a a = true.
Proof.
  intros a; apply eq_beq_true; trivial.
Qed.

Lemma beq_sym : forall a a', beq a a' = beq a' a.
Proof.
  intros a a'; destruct (eq_dec a a').
  subst a; trivial.
  rewrite (neq_beq_false n).
  rewrite (neq_beq_false (sym_not_eq n)).
  trivial.
Qed.

Ltac simplbeq :=
  match goal with 
    
    | H : beq ?r ?r = false |- _ =>
        rewrite (beq_refl r) in H;
          discriminate H; 
            simplbeq
        
    | H : context [ beq ?r ?r ] |- _ =>
        rewrite (beq_refl r) in H;
          simplbeq
   
    | |- context [ beq ?r ?r] =>
        rewrite (beq_refl r);
          simplbeq

    | H : beq ?rx ?ry = ?f
      |- context [beq ?rx ?ry] => 
        rewrite H; simplbeq
    | H : beq ?rx ?ry = ?f
      |- context [beq ?ry ?rx] => 
        rewrite (beq_sym rx ry) in H; 
          rewrite H; 
            simplbeq

    | H : beq ?rx ?ry = ?f,
      H1 : context [beq ?rx ?ry] |- _ =>
        rewrite H in H1; simplbeq
    | H : beq ?rx ?ry = ?f,
      H1 : context [beq ?ry ?rx] |- _ =>
        rewrite (beq_sym rx ry) in H; 
          rewrite H in H1;
            simplbeq

    | H : ?x = ?y 
      |- context [(beq ?x ?y)] => 
        rewrite (eq_beq_true H); simplbeq
    | H : ?y = ?x 
      |- context [(beq ?x ?y)] => 
        rewrite (eq_beq_true (sym_eq H));
          simplbeq
    | H : ?x = ?y,
      H1 : context [(beq ?x ?y)] |- _ => 
        rewrite (eq_beq_true H) in H1; simplbeq
    | H : ?y = ?x,
      H1 : context [(beq ?x ?y)] |- _ => 
        rewrite (eq_beq_true (sym_eq H)) in H1;
          simplbeq

    | H : ?x <> ?y 
      |- context [(beq ?x ?y)] => 
        rewrite (neq_beq_false H); simplbeq
    | H : ?y <> ?x 
      |- context [(beq ?x ?y)] => 
        rewrite (neq_beq_false (sym_not_eq H));
          simplbeq
    | H : ?x <> ?y,
      H1 : context [(beq ?x ?y)] |- _ => 
        rewrite (neq_beq_false H) in H1; simplbeq
    | H : ?y <> ?x,
      H1 : context [(beq ?x ?y)] |- _ => 
        rewrite (neq_beq_false (sym_not_eq H)) in H1;
          simplbeq

    | H : beq ?rx ?ry = true
      |- _ =>
        let Heq := fresh "Heq" in
        (generalize (beq_true_eq H); 
          intro Heq; subst rx); simplbeq

    | _ => idtac

  end.

(***********************)
(** get set            *)
(***********************)

Lemma set_a_get_a : forall (m : map) (a : A) (b : B),
  get (set m a b) a = SomeX b.
Proof.
  intros.
  assert (H:=set_sem m a a b).
  simplbeq.
  trivial.
Qed.
  
Lemma set_a_get_a' : forall (m : map) (a a' : A) (b : B),
  a <> a' -> get (set m a b) a' = get m a'.
Proof.
  intros.
  assert (H0:=set_sem m a a' b).
  simplbeq.
  trivial.
Qed.

(***********************)
(** lookup             *)
(***********************)

Lemma lookup_emp : forall (a : A) (b : B),
  lookup emp_map a b -> False.
Proof.
  intros a b; unfold lookup.
  intros H; rewrite (emp_sem a) in H.
  discriminate.
Qed.

Lemma lookup_get : forall (m : map) (a : A) (b : B),
  lookup m a b
  -> get m a = SomeX b.
Proof.
  intros m a b. unfold lookup. trivial.
Qed.

Lemma get_lookup : forall (m : map) (a : A) (b : B),
  get m a = SomeX b
    -> lookup m a b.
Proof.
  intros; unfold lookup; trivial.
Qed.

Lemma lookup_unique : forall (m : map) (a : A) (b b' :B),
  lookup m a b -> lookup m a b' -> b = b'.
Proof.
  intros m a b b'; unfold lookup.
  intros H1 H2; rewrite H1 in H2; injection H2.
  trivial.
Qed.

(**********************)
(** indom <--> lookup *)
(**********************)

Lemma lookup_indom : forall (m : map) (a : A) (b : B),
    lookup m a b -> indom m a.
Proof.
  intros m a b; unfold lookup, indom. 
  intros H; exists b; trivial.
Qed.

Lemma indom_lookup : forall (m : map) (a : A), 
  indom m a 
  -> exists b : B, 
    lookup m a b.
Proof.
  intros m a; unfold indom, lookup; trivial.
Qed.

(***********************)
(** indom <---> get    *)
(***********************)
 
Lemma indom_get : forall (m : map) (a : A),
  indom m a 
  -> exists b : B, 
    get m a = SomeX b.
Proof.
  intros m a; unfold indom; trivial.
Qed.

Lemma get_indom : forall (m : map) (a : A),
  (exists b : B, get m a = SomeX b)
  -> indom m a.
Proof.
  intros m a; unfold indom; trivial.
Qed.

Lemma nindom_get :
  forall (m : map) (a : A),
    ~ indom m a -> get m a = NoneX.
Proof.
  intros m a; unfold indom, not; trivial.
  destruct (get m a).
  intros H; elimtype False; apply H; exists b; trivial.
  trivial.
Qed.

(***********************)
(** sub                *)
(** meq                *)
(** compat             *)
(***********************)

Lemma sub_refl : forall (m : map),
  sub m m.
Proof.
  intros m a; destruct (get m a); trivial.
Qed.

(* prevent the funtion expanding from coq-type-checker *)
Implicit Arguments sub_refl[].

Lemma sub_asym : forall (m1 m2 : map),
  sub m1 m2 
  -> sub m2 m1
  -> meq m1 m2.
Proof.
  intros m1 m2; unfold  meq, sub, lookup.
  intros H1 H2; split; intros a b;
    substH H1 with (H1 a b); 
    substH H2 with (H2 a b);
    assumption.
Qed.

Lemma sub_trans : forall (m1 m2 m3 : map),
  sub m1 m2 
  -> sub m2 m3
  -> sub m1 m3.
Proof.
  intros m1 m2 m3; unfold sub.
  intros H1 H2 a b; 
    substH H1 with (H1 a b); 
    substH H2 with (H2 a b);
    auto.
Qed.
(* prevent the funtion expanding from coq-type-checker *)
Implicit Arguments sub_trans[m1 m2 m3].

Lemma meq_refl : forall (m : map), 
  meq m m.
Proof.
  intros m; unfold meq; split; apply sub_refl.
Qed.

Lemma meq_sym : forall (m1 m2 : map),
  meq m1 m2 -> meq m2 m1.
Proof.
  intros m1 m2; unfold meq.
  intros [H1 H2]; split; assumption.
Qed.

Lemma meq_trans : forall (m1 m2 m3 : map),
  meq m1 m2
  -> meq m2 m3
  -> meq m1 m3.
Proof.
  intros m1 m2 m3; unfold meq, sub.
  intros [H1 H2] [H3 H4];
    split; intros a b;
    substH H1 with (H1 a b);
    substH H2 with (H2 a b);
    substH H3 with (H3 a b);
    substH H4 with (H4 a b);
    tauto.
Qed.

Lemma meq_sem : forall (m1 m2 : map),
  meq m1 m2 -> 
  forall (a : A),
    match get m1 a, get m2 a with
      | SomeX b1, SomeX b2 => b1 = b2
      | NoneX, NoneX => True
      | _, _ => False
    end.
Proof.
  unfold meq, sub, lookup.
  intros m1 m2 [H1 H2] a.
  substH H1 with (H1 a).
  substH H2 with (H2 a).
  destruct (get m1 a);
    destruct (get m2 a);
      trivial.
  substH H2 with (H2 b0 (refl_equal _)).
  injection H2.
  trivial.
  substH H1 with (H1 b (refl_equal _)).
  discriminate.
  substH H2 with (H2 b (refl_equal _)).
  discriminate.
Qed.

Lemma compat_refl : forall m : map,
  compat m m -> compat m m.
Proof.
  unfold compat; intros m H a.
  substH H with (H a).
  destruct (get m a); auto.
Qed.

Lemma compat_sym : forall (m1 m2 : map),
  compat m1 m2 -> compat m2 m1.
Proof.
  unfold compat; intros m1 m2 H a.
  substH H with (H a).
  destruct (get m1 a); 
    destruct (get m2 a);
      auto.
Qed.
(***********************)
(** get    <--> sub    *)
(** lookup <--> sub    *)
(***********************)

Lemma get_sub_get : forall (m1 m2 : map) (a : A) (b : B),
  get m1 a = SomeX b -> sub m1 m2 -> get m2 a = SomeX b.
Proof.
  unfold sub.
  intros m1 m2 a b H H1.
  substH H1 with (H1 a b).
  apply (H1 H).
Qed.

Lemma lookup_sub_lookup : forall (m1 m2 : map) (a : A) (b : B),
  lookup m1 a b -> sub m1 m2 -> lookup m2 a b.
Proof.
  unfold sub.
  intros m1 m2 a b H H1.
  substH H1 with (H1 a b).
  apply (H1 H).
Qed.

Lemma indom_sub_indom : forall m1 m2 a, 
    indom m1 a -> sub m1 m2 -> indom m2 a.
Proof.
  unfold meq, sub. 
  intros m1 m2 a H H1. 
  substH H with (indom_lookup H).
  destruct H as [b H].
  substH H1 with (H1 a b).
  apply (lookup_indom (H1 H)).
Qed.

(***********************)
(** lookup <--> meq    *)
(***********************)

Lemma lookup_meq_lookup : forall (m1 m2 : map) (a : A) (b : B),
  lookup m1 a b -> meq m1 m2 -> lookup m2 a b.
Proof.
  unfold meq, sub.
  intros m1 m2 a b H0 [H1 H2].
  substH H1 with (H1 a b);
  substH H2 with (H2 a b).
  apply (H1 H0).
Qed.

Lemma indom_meq_indom : forall m1 m2 a, 
    indom m1 a -> meq m1 m2 -> indom m2 a.
Proof.
  unfold meq, sub. 
  intros m1 m2 a H [H1 H2]. 
  substH H with (indom_lookup H).
  destruct H as [b H].
  substH H1 with (H1 a b);
  substH H2 with (H2 a b).
  apply (lookup_indom (H1 H)).
Qed.

(************************)
(** disj                *)
(************************)

Lemma disj_emp_l : forall (m : map),
  disj (emp_map) m.
Proof.
  intros m a; rewrite (emp_sem a).
  destruct (get m a); trivial.
Qed.

Lemma disj_emp_r : forall (m : map),
  disj m (emp_map).
Proof.
  intros m a;rewrite (emp_sem a).
  destruct (get m a); trivial.
Qed.

Lemma disj_sym : forall (m1 m2 : map),
  disj m1 m2
  -> disj m2 m1.
Proof.
  intros m1 m2 H a;
    substH H with (H a);
    destruct (get m2 a); 
      destruct (get m1 a); trivial.
Qed.

(***********************)
(** indom <---> disj   *)
(***********************)

Lemma disj_indom : 
  forall (m1 m2 : map),
    disj m1 m2 
    -> (forall a : A, 
      (indom m1 a -> ~ indom m2 a)
      /\ (indom m2 a -> ~ indom m1 a)).
Proof.
  intros m1 m2 H a; unfold indom, not;
    substH H with (H a);split; 
      intros H1 H2; 
        destruct H1 as [b1 H1];
          destruct H2 as [b2 H2];
            destruct (get m1 a);
              destruct (get m2 a);
        discriminate || assumption.
Qed.

Lemma indom_disj : forall (m1 m2 : map),
  (forall a : A, 
    (indom m1 a -> ~ indom m2 a)
    /\ (indom m2 a -> ~ indom m1 a))
  -> disj m1 m2.
Proof.
  unfold indom, disj, not.
  intros m1 m2 H a.
    substH H with (H a).
  destruct H as [H1 H2]. 
      destruct (get m1 a);
      destruct (get m2 a);
        discriminate || trivial || (try assumption).
  apply H1.
  exists b; trivial.
  exists b0; trivial.
Qed.

(* (***********************) *)
(* (** disj --> lookup    *) *)
(* (***********************) *)
(* Lemma lookup_disj : forall (m1 m2 : map), *)
(*   (forall (a : A) (b : B), *)
(*     (lookup m1 a b -> ~ lookup m2 a b) *)
(*     /\ (lookup m1 a b -> ~ lookup m2 a b)) *)
(*   -> disj m1 m2. *)
(* Proof. *)
(*   intros m1 m2 H. *)
(*   unfold disj. *)
(*   intro a. *)
(* Qed. *)

(* Lemma disj_lookup forall (m1 m2 : map) (a : A) (b : B), *)
(*   disj m1 m2 *)
(*   ->   *)
(*   (forall (a : A) (b : B), *)
(*     (lookup m1 a b -> ~ lookup m2 a b) *)
(*     /\ (lookup m1 a b -> ~ lookup m2 a b)) *)
(* Proof. *)

(* Qed. *)

(***********************)
(** disj <--> meq       *)
(***********************)


Lemma disj_meq_disj_r : forall (m1 m2 m3 : map),
  disj m1 m2 
  -> meq m2 m3 
  -> disj m1 m3.
Proof.
  unfold disj, meq, sub, lookup.
  intros m1 m2 m3 Hj [H1 H2] a.
  substH Hj with (Hj a).
  substH H1 with (H1 a).
  substH H2 with (H2 a).
  destruct (get m1 a);
    destruct (get m2 a);
      destruct (get m3 a);
        assumption || trivial.
  substH H2 with (H2 b0 (refl_equal _)); discriminate.
Qed.

Lemma disj_meq_disj_l : forall (m1 m2 m3 : map),
  disj m1 m2 
  -> meq m1 m3 
  -> disj m3 m2.
Proof.
  unfold disj, meq, sub, lookup.
  intros m1 m2 m3 Hj [H1 H2] a.
  substH Hj with (Hj a).
  substH H1 with (H1 a).
  substH H2 with (H2 a).
  destruct (get m1 a);
    destruct (get m2 a);
      destruct (get m3 a);
        assumption || trivial.
  substH H2 with (H2 b0 (refl_equal _)); discriminate.
Qed.

(***********************)
(** disj --> compat    *)
(***********************)

Lemma disj_compat : forall m1 m2 : map,
  disj m1 m2 -> compat m1 m2.
Proof.
  unfold disj, compat.
  intros m1 m2 H a.
  substH H with (H a).
  destruct (get m1 a); 
    destruct (get m2 a);
      auto.
  discri.
Qed.

(***********************)
(** sub <--> compat    *)
(***********************)

Lemma compat_sub_l : forall (m1 m2 : map),
  compat m1 m2 -> sub m1 (merge m1 m2).
Proof.
  intros m1 m2.
  unfold compat, sub, lookup.
  intros H a b.
  substH H with (H a).
  rewrite (merge_sem m1 m2 a).
  destruct (get m1 a); destruct (get m2 a); 
    auto.
  discri.
Qed.

Lemma compat_sub_r : forall (m1 m2 : map),
    compat m1 m2 -> sub m2 (merge m1 m2).
Proof.
  intros m1 m2.
  unfold compat, sub, lookup.
  intros H a b.
  substH H with (H a).
  rewrite (merge_sem m1 m2 a).
  destruct (get m1 a); destruct (get m2 a); 
    auto.
  subst b0; auto.
  discri.
Qed.

Lemma sub_compat :
  forall (m1 m2 : map),
    sub m1 m2 -> compat m1 m2.
Proof.
  unfold sub, compat, lookup.
  intros m1 m2 H a.
  substH H with (H a).
  destruct (get m1 a); destruct (get m2 a);
    try tauto; auto.
  substH H with (H b (refl_equal _)). 
  injection H; auto.
Qed.

(*************************************************)
(**                                              *)
(** *  merge (union)                             *)
(**                                              *)
(*************************************************)

(* 
  sub <--> merge
*)

Lemma sub_merge_sub_l :
  forall (m1 m2 : map),
    compat m1 m2
    -> sub m1 (merge m1 m2).
Proof.
  intros m1 m2; unfold compat, sub.
  intros H a b.
  unfold lookup.
  rewrite (merge_sem m1 m2 a).
  substH H with (H a).
  destruct (get m1 a);
    destruct (get m2 a);
      auto; discri.
Qed.

Lemma sub_merge_sub_r :
  forall (m1 m2 : map),
    compat m1 m2
    -> sub m2 (merge m1 m2).
Proof.
  intros m1 m2; unfold compat, sub.
  intros H a b.
  unfold lookup.
  rewrite (merge_sem m1 m2 a).
  substH H with (H a).
  destruct (get m1 a);
    destruct (get m2 a);
      auto.
  subst b0. trivial.
  discri.
Qed.

Lemma sub_merge_elim_l :
  forall (m1 m2 m3: map),
    compat m1 m2
    -> sub (merge m1 m2) m3
    -> sub m1 m3.
Proof.
  unfold compat, sub, lookup.
  intros m1 m2 m3 Hcomp Hsub a b.
  substH Hsub with (Hsub a b).
  substH Hcomp with (Hcomp a).
  rewrite (merge_sem m1 m2 a) in Hsub.
  destruct (get m1 a);
    destruct (get m2 a);
      auto; discri.
Qed.
Implicit Arguments sub_merge_elim_l [m1 m2 m3].

Lemma sub_merge_elim_r :
  forall (m1 m2 m3: map),
    compat m1 m2
    -> sub (merge m1 m2) m3
    -> sub m2 m3.
Proof.
  unfold compat, sub, lookup.
  intros m1 m2 m3 Hcomp Hsub a b.
  substH Hsub with (Hsub a b).
  substH Hcomp with (Hcomp a).
  rewrite (merge_sem m1 m2 a) in Hsub.
  destruct (get m1 a);
    destruct (get m2 a);
      auto; try discri.
  subst b0. trivial.
Qed.
Implicit Arguments sub_merge_elim_r [m1 m2 m3].

Lemma sub_merge_intro_l :
  forall (m1 m2 m3: map),
    compat m1 m2
    -> sub m3 m1
    -> sub m3 (merge m1 m2).
Proof.
  unfold compat, sub, lookup.
  intros m1 m2 m3 Hcomp Hsub a b.
  substH Hsub with (Hsub a b).
  substH Hcomp with (Hcomp a).
  rewrite (merge_sem m1 m2 a).
  destruct (get m1 a);
    destruct (get m2 a);
      auto; try discri.
  intros H.
  assert (Hf := Hsub H); discriminate.
Qed.
Implicit Arguments sub_merge_intro_l [m1 m2 m3].

Lemma sub_merge_intro_r :
  forall (m1 m2 m3: map),
    compat m1 m2
    -> sub m3 m2
    -> sub m3 (merge m1 m2).
Proof.
  unfold compat, sub, lookup.
  intros m1 m2 m3 Hcomp Hsub a b.
  substH Hsub with (Hsub a b).
  substH Hcomp with (Hcomp a).
  rewrite (merge_sem m1 m2 a).
  destruct (get m1 a);
    destruct (get m2 a);
      auto; try discri.
  subst b0. trivial.
  intros H.
  assert (Hf := Hsub H); discriminate.
Qed.
Implicit Arguments sub_merge_intro_r [m1 m2 m3].


Lemma merge_sub_intro : forall (m m1 m2 : map),
  sub m1 m 
  -> sub m2 m
  -> sub (merge m1 m2) m.
Proof.
  unfold sub, lookup; intros m m1 m2 H1 H2.
  intros a b.
  substH H1 with (H1 a b).
  substH H2 with (H2 a b).
  rewrite (merge_sem m1 m2 a).
  intros H.
  destruct (get m1 a); destruct (get m2 a);
    destruct (get m a); try tauto; try auto.
Qed.
Implicit Arguments merge_sub_intro [m m1 m2].

(*
  meq <--> merge
*)

Lemma meq_merge_sym : forall (m1 m2 : map),
  compat m1 m2
  -> meq (merge m1 m2) (merge m2 m1).
Proof.
  unfold compat, meq, sub, lookup.
  intros m1 m2 H; split; intros a b;
    substH H with (H a);
    rewrite (merge_sem m1 m2 a);
      rewrite (merge_sem m2 m1 a);
        destruct (get m1 a);
          destruct (get m2 a);
            auto; try contradiction.
  subst b0; trivial.
  subst b0; trivial.
Qed.

Lemma meq_merge_assoc_l : forall (m1 m2 m3 : map),
  meq 
  (merge (merge m1 m2) m3)
  (merge m1 (merge m2 m3)).
Proof.
  unfold disj, meq, sub, lookup.
  intros m1 m2 m3; split; intros a b;
    rewrite (merge_sem m1 (merge m2 m3) a);
    rewrite (merge_sem (merge m1 m2) m3 a);
    rewrite (merge_sem m1 m2 a);
    rewrite (merge_sem m2 m3 a);
      destruct (get m1 a);
        destruct (get m2 a);
        destruct (get m3 a);
          trivial.
Qed.

Lemma meq_merge_assoc_r : forall (m1 m2 m3 : map),
  meq 
  (merge m1 (merge m2 m3))
  (merge (merge m1 m2) m3).
Proof.
  unfold disj, meq, sub, lookup.
  intros m1 m2 m3; split; intros a b;
    rewrite (merge_sem m1 (merge m2 m3) a);
    rewrite (merge_sem (merge m1 m2) m3 a);
    rewrite (merge_sem m1 m2 a);
    rewrite (merge_sem m2 m3 a);
      destruct (get m1 a);
        destruct (get m2 a);
        destruct (get m3 a);
          trivial.
Qed.

Lemma meq_merge_emp_l : forall (m : map),
  meq (merge emp_map m) m.
Proof.
  unfold meq, sub, lookup.
  intros m; split; intros a b;
    rewrite (merge_sem emp_map m a);
      rewrite (emp_sem a);
        destruct (get m a); 
          trivial.
Qed.

Lemma meq_merge_emp_r : forall (m : map),
  meq (merge m emp_map) m.
Proof.
  unfold meq, sub, lookup.
  intros m; split; intros a b;
    rewrite (merge_sem m emp_map a);
      rewrite (emp_sem a);
        destruct (get m a); 
          trivial.
Qed.

Lemma meq_merge_meq : 
  forall m1 m2 m3 m4,
    meq m1 m3 -> meq m2 m4
    -> meq (merge m1 m2) (merge m3 m4).
Proof.
  unfold meq, sub, lookup.
  intros m1 m2 m3 m4 [H1 H2] [H3 H4].
  split; intros a b;
    substH H1 with (H1 a);
    substH H2 with (H2 a);
    substH H3 with (H3 a);
    substH H4 with (H4 a);
    rewrite (merge_sem m1 m2 a);
    rewrite (merge_sem m3 m4 a);
    intros H;
    destruct (get m1 a) as [b1 |];
    destruct (get m2 a) as [b2 |];
    destruct (get m3 a) as [b3 |];
    destruct (get m4 a) as [b4 |];
      auto.
  assert (Hf := H1 b H); discriminate.
  assert (Hf := H2 b3 (refl_equal _)); discriminate.
  assert (Hf := H1 b1 (refl_equal _)); discriminate.
  assert (Hf := H2 b H); discriminate.
Qed.

(************************)
(** lookup <--> merge   *)
(************************)

Lemma lookup_merge_elim : 
  forall (m1 m2 : map) (a : A) (b : B),
    lookup (merge m1 m2) a b 
    -> lookup m1 a b \/ lookup m2 a b.
Proof.
  intros m1 m2 a b.
  unfold lookup; rewrite (merge_sem m1 m2 a).
  destruct (get m1 a); 
    destruct (get m2 a); 
      auto.
Qed.

Lemma lookup_merge_intro :
  forall (m1 m2 : map) (a : A) (b : B), 
    lookup m1 a b \/ lookup m2 a b
    -> compat m1 m2
    -> lookup (merge m1 m2) a b.
Proof.
  intros m1 m2 a b.
  unfold lookup, compat; rewrite (merge_sem m1 m2 a).
  intros [H1 | H2]; intros H; 
    substH H with (H a);   
    destruct (get m1 a); 
      destruct (get m2 a); 
        discriminate || contradiction || auto.
  subst b0; trivial.
Qed.

(********************************)
(** indom <--> merge            *)
(********************************)

Lemma indom_merge_intro :
  forall (m1 m2 : map) (a : A),
    indom m1 a \/ indom m2 a
    -> indom (merge m1 m2) a.
Proof. 
  (* < *)
  unfold indom.
  intros m1 m2 a [[b H]|[b H]];
    rewrite (merge_sem m1 m2 a);
        destruct (get m2 a);
          destruct (get m1 a); 
            discriminate || trivial.
  exists b1; trivial.
  exists b0; trivial.
  exists b1; trivial.
  exists b0; trivial.
  (* > *)
Qed.

Lemma indom_merge_elim : 
  forall (m1 m2 : map) (a : A),
  indom (merge m1 m2) a
  -> indom m1 a \/ indom m2 a.
Proof.
  (* < *)
  unfold indom.
  intros m1 m2 a [b H].
  rewrite (merge_sem m1 m2 a) in H.
  destruct (get m2 a);
    destruct (get m1 a).
  left; exists b; assumption.
  right; exists b; assumption.
  left; exists b; assumption.
  right; exists b; assumption.
  (* > *)
Qed.

(********************************)
(** disj <--> disj merge        *)
(********************************)

Lemma disj_merge_intro_l :
  forall (m1 m2 m3 : map),
    disj m1 m3 /\ disj m2 m3
    -> disj (merge m1 m2) m3.
Proof.
  unfold disj.
  intros m1 m2 m3 [H1 H2] a.
  substH H1 with (H1 a).
  substH H2 with (H2 a).
  rewrite (merge_sem m1 m2 a);
    destruct (get m2 a);
      destruct (get m1 a); 
        discriminate || trivial.
Qed.

Lemma disj_merge_elim_l :
  forall (m1 m2 m3 : map),
    disj (merge m1 m2) m3    
    -> disj m1 m3 /\ disj m2 m3.
Proof.
  unfold disj.
  intros m1 m2 m3 H.
  split; intro a; substH H with (H a);
    rewrite (merge_sem m1 m2 a) in H;
      destruct (get m3 a);
        destruct (get m2 a);
          destruct (get m1 a); 
            discriminate || trivial.
Qed.

Lemma disj_merge_intro_r :
  forall (m1 m2 m3 : map),
    disj m1 m2 /\ disj m1 m3
    -> disj m1 (merge m2 m3).
Proof.
  unfold disj.
  intros m1 m2 m3 [H1 H2] a.
  substH H1 with (H1 a).
  substH H2 with (H2 a).
  rewrite (merge_sem m2 m3 a);
    destruct (get m2 a);
      destruct (get m3 a); 
        discriminate || trivial.
Qed.

Lemma disj_merge_elim_r :
  forall (m1 m2 m3 : map),
    disj m1 (merge m2 m3)
    -> disj m1 m2 /\ disj m1 m3.
Proof.
  unfold disj.
  intros m1 m2 m3 H.
  split; intro a; substH H with (H a);
    rewrite (merge_sem m2 m3 a) in H;
      destruct (get m3 a);
        destruct (get m2 a);
          destruct (get m1 a); 
            discriminate || trivial.
Qed.

Lemma disj_merge_intro_lr :
  forall (m1 m2 m1' m2' : map),
    disj m1 m1'
    -> disj m1 m2'
    -> disj m2 m1'
    -> disj m2 m2'
    -> disj (merge m1 m2) (merge m1' m2').
Proof.
  unfold disj.
  intros m1 m2 m1' m2' H1 H2 H3 H4 a.
  substH H1 with (H1 a);
  substH H2 with (H2 a);
  substH H3 with (H3 a);
  substH H4 with (H4 a).
  rewrite (merge_sem m1 m2 a);
    rewrite (merge_sem m1' m2' a).
  destruct (get m1 a);
    destruct (get m2 a); 
      destruct (get m1' a);
        destruct (get m2' a); 
          discriminate || trivial.
Qed.

Lemma disj_merge_elim_lr :
  forall (m1 m2 m1' m2' : map),
    disj (merge m1' m2') (merge m1 m2)
    -> (disj m1 m1' /\ disj m1 m2')
    /\ (disj m2 m1' /\ disj m2 m2').
Proof.
  unfold disj.
  intros m1 m2 m1' m2' H.
  split; split; intro a;
    substH H with (H a);
      rewrite (merge_sem m1 m2 a) in H;
        rewrite (merge_sem m1' m2' a) in H;
          destruct (get m1 a);
            destruct (get m2 a); 
              destruct (get m1' a);
                destruct (get m2' a); 
                  discriminate || trivial.
Qed.

(*************************************************)
(**                                              *)
(** * join                                   *)
(**                                              *)
(*************************************************)


Lemma join_comm :
  forall ma mb mab : map,
    join ma mb mab -> join mb ma mab.
Proof.
  intros ma mb mab.
  unfold join.
  intros H a.
  substH H with (H a).
  destruct (get ma a);
    destruct (get mb a);
      destruct (get mab a);
        auto.
Qed.

Lemma join_assoc_l :
  forall ma mb mc mab mabc,
    join ma mb mab->
    join mab mc mabc ->
    exists mbc, join mb mc mbc /\
      join ma mbc mabc.
Proof.
  intros ma mb mc mab mabc. 
  unfold join.
  intros H1 H2.
  exists (merge mb mc).
  split; intros a; 
    rewrite (merge_sem mb mc a); 
    substH H1 with (H1 a);
    substH H2 with (H2 a);
    destruct (get ma a);
      destruct (get mb a);
        destruct (get mc a);
          destruct (get mab a);
            destruct (get mabc a); 
              try tauto; try subst b; auto.
Qed.

Lemma join_assoc_r :
  forall ma mb mc mbc mabc,
    join mb mc mbc->
    join ma mbc mabc ->
    exists mab, join ma mb mab /\
      join mab mc mabc.
Proof.
  intros ma mb mc mbc mabc. 
  unfold join.
  intros H1 H2.
  exists (merge ma mb).
  split; intros a;
    rewrite (merge_sem ma mb a);
    substH H1 with (H1 a);
    substH H2 with (H2 a);
    destruct (get ma a);
      destruct (get mb a);
        destruct (get mc a);
          destruct (get mbc a);
            destruct (get mabc a); 
              try tauto; try subst b; auto.
Qed.

Lemma join_disj_meq :
  forall (m1 m2 m : map),
    join m1 m2 m
    -> disj m1 m2 /\ meq (merge m1 m2) m.
Proof.
  unfold join, disj, meq,sub, lookup.
  intros m1 m2 m Hj; split. 
  intros a; substH Hj with (Hj a).
  destruct (get m1 a);
  destruct (get m2 a);
  destruct (get m a);
    trivial.
  split; intros a b;
  substH Hj with (Hj a);
  rewrite (merge_sem m1 m2 a);
    destruct (get m1 a);
      destruct (get m2 a); 
        destruct (get m a); 
        (contradiction || (try subst b0)); trivial.
Qed.

Lemma disj_meq_join :
  forall (m1 m2 m: map),
    disj m1 m2 /\ meq (merge m1 m2) m
    -> join m1 m2 m.
Proof.
  unfold disj, meq,sub, lookup.
  intros m1 m2 m [H0 [H1 H2]] a.
  substH H0 with (H0 a);
  substH H1 with (H1 a);
  substH H2 with (H2 a).
  rewrite (merge_sem m1 m2 a) in H1.
  rewrite (merge_sem m1 m2 a) in H2.
  destruct (get m a);
    destruct (get m1 a);
      destruct (get m2 a); 
        contradiction 
          || (try substH H2 
            with (H2 b (refl_equal _)); 
          (injection H2 || discriminate || trivial; 
            trivial))
          || (try substH H2 
            with (H1 b (refl_equal _)); 
          (injection H1 || discriminate || trivial; 
            trivial));
          trivial.
Qed.

Lemma join_sub_l :
  forall m1 m2 m,
    join m1 m2 m
    -> sub m1 m.
Proof.
  intros m1 m2 m.
  unfold join, sub, lookup; intros H.
  intro a.
  substH H with (H a).
  destruct (get m1 a);
    destruct (get m2 a);
        destruct (get m a); 
          tauto || auto.
  subst; trivial.
  intros; discriminate.
Qed.

Lemma join_sub_r :
  forall m1 m2 m,
    join m1 m2 m
    -> sub m2 m.
Proof.
  intros m1 m2 m.
  unfold join, sub, lookup; intros H.
  intro a.
  substH H with (H a).
  destruct (get m1 a);
    destruct (get m2 a);
        destruct (get m a); 
          tauto || auto.
  intros; discriminate.
  subst; trivial.
Qed.
Implicit Arguments join_sub_r [m1 m2 m].

Lemma join_lookup_lookup_l :
  forall m1 m2 m a b,
    join m1 m2 m
    -> lookup m1 a b
    -> lookup m a b.
Proof.
  intros m1 m2 m a b Hj H1.
  assert (H:= join_sub_l Hj).
  unfold sub in H.
  apply (H a b H1).
Qed.

Lemma join_lookup_lookup_r :
  forall m1 m2 m a b,
    join m1 m2 m
    -> lookup m2 a b
    -> lookup m a b.
Proof.
  intros m1 m2 m a b Hj H2.
  assert (H:= join_sub_r Hj).
  unfold sub in H.
  apply (H a b H2).
Qed.

Lemma join_set_l :
  forall m1 m2 m a b,
    join m1 m2 m
    -> indom m1 a
    -> join (set m1 a b) m2 (set m a b).
Proof.
  unfold indom, join.
  intros m1 m2 m a b H H1.
  intro a0.
  substH H with (H a0).
  destruct (eq_dec a a0) as [Heq | Hneq].
  subst a0.
  rewrite (set_a_get_a m1 a b).
  rewrite (set_a_get_a m a b).
  destruct (get m1 a);
    destruct (get m2 a);
      destruct (get m a);
        tauto || auto.
  destruct H1.
  discriminate.
  rewrite (set_a_get_a' m1 b Hneq).
  rewrite (set_a_get_a' m b Hneq).
  destruct (get m1 a0);
    destruct (get m2 a0);
      destruct (get m a0);
        tauto || auto.
Qed.

Lemma join_set_r :
  forall m1 m2 m a b,
    join m1 m2 m
    -> indom m2 a
    -> join m1 (set m2 a b) (set m a b).
Proof.
  unfold indom, join.
  intros m1 m2 m a b H H1.
  intro a0.
  substH H with (H a0).
  destruct (eq_dec a a0) as [Heq | Hneq].
  subst a0.
  rewrite (set_a_get_a m2 a b).
  rewrite (set_a_get_a m a b).
  destruct (get m1 a);
    destruct (get m2 a);
      destruct (get m a);
        tauto || auto.
  destruct H1.
  discriminate.
  rewrite (set_a_get_a' m2 b Hneq).
  rewrite (set_a_get_a' m b Hneq).
  destruct (get m1 a0);
    destruct (get m2 a0);
      destruct (get m a0);
        tauto || auto.
Qed.

(***********************************************************)
(**                                                        *)
(**    Tactics                                             *)
(**                                                        *)
(***********************************************************)

Ltac simplgs := 
  match goal with
    | [ |- context [(get (set ?m ?a ?b) ?a)] ] =>
      rewrite (set_a_get_a m a b); simplgs

    | [ H : ?a <> ?a'
      |- context [(get (set ?m ?a ?b) ?a')] ] =>
      rewrite (@set_a_get_a' m a a' b H); simplgs

    | [ H : ?a' <> ?a
      |- context [(get (set ?m ?a ?b) ?a')] ] =>
      rewrite (@set_a_get_a' m a a' b (sym_not_eq H)); simplgs

    | [ |- context [(get (set ?m ?a ?b) ?a')] ] =>
      rewrite (@set_a_get_a' m a a' b); simplgs

    | [ |- context [(lookup (set ?m ?a ?b) ?a' ?b')] ] =>
      apply get_lookup; 
        rewrite (@set_a_get_a' m a a' b); simplgs

    | _ => idtac
  end.
