(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Tiny Machine                                                *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: codeheap.v,v 1.4 2006/07/15 02:56:42 ssg05 Exp $ *)

Require regfile.
Require word.
Require label.
Require Import instr.

Require Export Coq.Lists.List.
Require Import MAP.
Require Import LMapB.

Module Instr <: ONE_SET.
  Definition B := instr.
End Instr.
Module CHeapDef := ListMap Instr.
Module CHeap := MapEx CHeapDef.

Import label.
Import instr.
Import CHeap.
Import reg.

Definition codeheap := CHeap.map.

Definition get : codeheap -> label -> option instr 
  := CHeap.get.

Inductive getis : codeheap -> label -> instrseq -> Prop:=
  | getis_j : 
    forall (ch : codeheap) (f f': label),
      (get ch f = Some (j f')) 
      -> getis ch f (ione (j f'))
  | getis_jr : 
    forall (ch : codeheap) (f : label) (r : reg),
      (get ch f = Some (jr r)) 
      -> getis ch f (ione (jr r))
  | getis_seq :
    forall (ch : codeheap) (f : label) 
      (i : instr) (is : instrseq),
      (get ch f = Some i) -> getis ch (f+1) is 
      -> getis ch f (iseq i is).

Lemma getis_merge_intro :
  forall (m1 m2 : codeheap) (f : label) (is : instrseq), 
    getis m1 f is \/ getis m2 f is
    -> disj m1 m2
    -> getis (merge m1 m2) f is.
Proof.
  intros m1 m2 f is His Hdisj.
  destruct His as [His | His].
  induction His as 
    [ m1 f f' H
      | m1 f r H
      | m1 f i is H His IHis
    ].
  apply getis_j.
    unfold get in * .
    rewrite (merge_sem m1 m2 f). 
    rewrite H.
    destruct (CHeap.get m2 f); trivial.
  apply getis_jr.
    unfold get in * .
    rewrite (merge_sem m1 m2 f). 
    rewrite H.
    destruct (CHeap.get m2 f); trivial.
  apply getis_seq.
    unfold get in * .
    rewrite (merge_sem m1 m2 f). 
    rewrite H.
    destruct (CHeap.get m2 f); trivial.
    apply (IHis Hdisj).

  induction His as 
    [ m2 f f' H
      | m2 f r H
      | m2 f i is H His IHis
    ].
  apply getis_j.
    unfold get in * .
    assert (Hdisj' := Hdisj f).
    rewrite (merge_sem m1 m2 f). 
    destruct (CHeap.get m1 f); trivial.
    rewrite H in Hdisj'; contradiction.
    rewrite H; trivial.
  apply getis_jr.
    unfold get in * .
    assert (Hdisj' := Hdisj f).
    rewrite (merge_sem m1 m2 f). 
    destruct (CHeap.get m1 f); trivial.
    rewrite H in Hdisj'; contradiction.
    rewrite H; trivial.
  apply getis_seq.
    unfold get in * .
    assert (Hdisj' := Hdisj f).
    rewrite (merge_sem m1 m2 f). 
    destruct (CHeap.get m1 f); trivial.
    rewrite H in Hdisj'; contradiction.
    rewrite H; trivial.
    apply (IHis Hdisj).
Qed.

Lemma getis_compat_merge_intro :
  forall (m1 m2 : codeheap) (f : label) (is : instrseq), 
    getis m1 f is \/ getis m2 f is
    -> compat m1 m2
    -> getis (merge m1 m2) f is.
Proof.
  intros m1 m2 f is His Hcomp.
  destruct His as [His | His].
  induction His as 
    [ m1 f f' H
      | m1 f r H
      | m1 f i is H His IHis
    ].
  apply getis_j.
    unfold get in * .
    rewrite (merge_sem m1 m2 f). 
    rewrite H.
    destruct (CHeap.get m2 f); trivial.
  apply getis_jr.
    unfold get in * .
    rewrite (merge_sem m1 m2 f). 
    rewrite H.
    destruct (CHeap.get m2 f); trivial.
  apply getis_seq.
    unfold get in * .
    rewrite (merge_sem m1 m2 f). 
    rewrite H.
    destruct (CHeap.get m2 f); trivial.
    apply (IHis Hcomp).

  induction His as 
    [ m2 f f' H
      | m2 f r H
      | m2 f i is H His IHis
    ].
  apply getis_j.
    unfold get in * .
    assert (Hcomp' := Hcomp f).
    rewrite (merge_sem m1 m2 f). 
    destruct (CHeap.get m1 f); trivial.
    rewrite H in Hcomp'. 
    subst b.
    destruct (CHeap.get m2 f); trivial.
    rewrite H; trivial.
  apply getis_jr.
    unfold get in * .
    assert (Hcomp' := Hcomp f).
    rewrite (merge_sem m1 m2 f). 
    destruct (CHeap.get m1 f); trivial.
    rewrite H in Hcomp'. 
    subst b.
    destruct (CHeap.get m2 f); trivial.
    rewrite H; trivial.
  apply getis_seq.
    unfold get in * .
    assert (Hcomp' := Hcomp f).
    rewrite (merge_sem m1 m2 f). 
    destruct (CHeap.get m1 f); trivial.
    rewrite H in Hcomp'. 
    subst b.
    destruct (CHeap.get m2 f); trivial.
    rewrite H; trivial.
    apply (IHis Hcomp).
Qed.

Ltac getis_solve :=
  match goal with
    | |- getis ?C ?l (iseq ?i ?is) =>
      constructor; 
        [compute; simpl; reflexivity | getis_solve]
    | |- getis ?C ?l (ione ?i) =>
      constructor; compute; simpl; reflexivity
    | |- getis ?C ?l ?code =>
      unfold code; getis_solve
    | _ => idtac
  end.

