(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Register Definition                                         *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: reg.v,v 1.4 2006/07/15 02:56:42 ssg05 Exp $ *)

Inductive reg : Set :=
  | r0 : reg
  | r1 : reg
  | r2 : reg
  | r3 : reg
  | r4 : reg
  | r5 : reg
  | r6 : reg
  | r7 : reg
  | r8 : reg
  | r9 : reg
  | r10 : reg
  | r11 : reg
  | r12 : reg
  | r13 : reg
  | r14 : reg
  | r15 : reg
  | r16 : reg
  | r17 : reg
  | r18 : reg
  | r19 : reg
  | r20 : reg
  | r21 : reg
  | r22 : reg
  | r23 : reg
  | r24 : reg
  | r25 : reg
  | r26 : reg
  | r27 : reg
  | r28 : reg
  | r29 : reg
  | r30 : reg
  | r31 : reg.

Implicit Types r : reg.

Definition beq_reg (r r' : reg) :=
  match r, r' with
  | r0, r0 => true
  | r1, r1 => true
  | r2, r2 => true
  | r3, r3 => true
  | r4, r4 => true
  | r5, r5 => true
  | r6, r6 => true
  | r7, r7 => true
  | r8, r8 => true
  | r9, r9 => true
  | r10, r10 => true
  | r11, r11 => true
  | r12, r12 => true
  | r13, r13 => true
  | r14, r14 => true
  | r15, r15 => true
  | r16, r16 => true
  | r17, r17 => true
  | r18, r18 => true
  | r19, r19 => true
  | r20, r20 => true
  | r21, r21 => true
  | r22, r22 => true
  | r23, r23 => true
  | r24, r24 => true
  | r25, r25 => true
  | r26, r26 => true
  | r27, r27 => true
  | r28, r28 => true
  | r29, r29 => true
  | r30, r30 => true
  | r31, r31 => true
  | _, _ => false
  end.

Lemma beq_sym : forall r r', beq_reg r r' = beq_reg r' r.
Proof.
  double induction r r'; simpl; intros; auto.
Qed.

Lemma beq_refl : forall r , beq_reg r r = true.
Proof.
  induction r; simpl; intros; auto.
Qed.

Lemma beq_req_true : forall r r', r = r' -> beq_reg r r' = true.
  intros.
  rewrite H.
  induction r';auto.
Qed.

Lemma beq_true_req : forall r r', beq_reg r r' = true -> r = r'.
Proof.
  double induction r r'; simpl; intros; discriminate || trivial.
Qed.

(* Lemma beq_false_rneq : forall r r', beq_reg r r' = true -> r = r'. *)
(* Proof. *)
(*   double induction r r'; simpl; intros; discriminate || trivial. *)
(* Qed. *)

Lemma beq_rneq_false : forall r r', r <> r' -> beq_reg r r' = false.
  intros.
  induction r;
    induction r';
      tauto.
Qed.

Lemma eq_dec : forall r r', {r=r'}+{r<>r'}.
Proof.
  intros r r'; compare r r';
  intro; [left | right]; assumption.
Qed.

Lemma beq_dec : forall r r', 
  {beq_reg r r' = true} + {beq_reg r r' = false}.
Proof.
  double induction r r'; simpl; intros; auto.
Qed.

(* MIPS register conventions *)
   (* from Xinyu's scap.v *)
Notation ra := r31.  (* return address proc *)
Notation zero := r0. (*always zero *)
Notation ast := r1.  (* assembler temporary *)
Notation v0 := r2.   (*v0 - v1 return values *)
Notation v1 := r3.
Notation a0 := r4.   (*a0 - a3 arguments *)
Notation a1 := r5.
Notation a2 := r6.
Notation a3 := r7.
Notation t0 := r8.   (*t0 - t9 caller saved temporary *)
Notation t1 := r9.
Notation t2 := r10. 
Notation t3 := r11.
Notation t4 := r12.
Notation t5 := r13.
Notation t6 := r14.
Notation t7 := r15.
Notation t8 := r24.
Notation t9 := r25.
Notation s0 := r16.  (*s0 - s7 callee saved temporary *)
Notation s1 := r17.
Notation s2 := r18.
Notation s3 := r19.
Notation s4 := r20.
Notation s5 := r21.
Notation s6 := r22.
Notation s7 := r23.
Notation k0 := r26.
Notation k1 := r27.
Notation gp := r28.
Notation sp := r29.
Notation fp := r30.

Ltac breg_simpl :=
  match goal with 
    
    | H : beq_reg ?r ?r = false |- _ =>
        rewrite (beq_req_true r r (refl_equal r)) in H;
          discriminate H; 
            breg_simpl
        
    | H : context [ beq_reg ?r ?r = true] |- _ =>
        rewrite (beq_req_true r r (refl_equal r)) in H;
          breg_simpl
   
    | |- context [ beq_reg ?r ?r] =>
        rewrite (beq_req_true r r (refl_equal r));
          breg_simpl

    | H : beq_reg ?rx ?ry = ?f
      |- context [beq_reg ?rx ?ry] => 
        rewrite H; breg_simpl
    | H : beq_reg ?rx ?ry = ?f
      |- context [beq_reg ?ry ?rx] => 
        rewrite (beq_sym rx ry) in H; 
          rewrite H; 
            breg_simpl

    | H : beq_reg ?rx ?ry = ?f,
      H1 : context [beq_reg ?rx ?ry] |- _ =>
        rewrite H in H1; breg_simpl
    | H : beq_reg ?rx ?ry = ?f,
      H1 : context [beq_reg ?ry ?rx] |- _ =>
        rewrite (beq_sym rx ry) in H; 
          rewrite H in H1;
            breg_simpl

    | H : beq_reg ?rx ?ry = true
      |- _ =>
        let Heq := fresh "Heq" in
        (generalize (beq_true_req rx ry H); 
          intro Heq; subst rx); breg_simpl

    | _ => idtac

  end.

Ltac breg_solve :=
  intros; 
  match goal with 
    | |- context [ beq_reg ?r ?r] =>
        rewrite (beq_req_true r r (refl_equal r));
          reflexivity

    | H : beq_reg ?rx ?ry = ?f 
      |- beq_reg ?rx ?ry = ?f =>
        assumption

    | H : beq_reg ?rx ?ry = ?f 
      |- beq_reg ?ry ?rx = ?f =>
        rewrite (beq_sym rx ry) in H;  assumption
    
    | H : beq_reg ?rx ?ry = true
      |- _ =>
        let Heq := fresh "Heq" in
        (generalize (beq_true_req rx ry H); 
          intro Heq; subst rx);  breg_solve
        
    | H : beq_reg ?r ?r = false
      |- _ =>
        rewrite (beq_req_true r r (refl_equal r)) in H;
          discriminate H
            
    | |- beq_reg ?rx ?ry = ?f =>
        simpl; reflexivity

    | _ => fail 1 "The goal cannot be solved. "
end.

Section test.

  Variables rx ry rz r r' : reg.
  Variables ra rb rc rd re : reg.

  Hypotheses 
    (H0 : beq_reg rx ry = false)
(*     (H1 : beq_reg ry rz = true) *)
(*     (H2 : beq_reg rz r = true) *)
(*     (H3 : beq_reg ra rb = true) *)
(*     (H4 : beq_reg ra rc = true) *)
(*     (H5 : beq_reg rd ra = true). *)
.
(* Goal beq_reg ry rx = false. *)
(* breg_solve. *)

(* Goal beq_reg rx r = true. breg_solve. *)

(* Goal beq_reg rx r = false -> beq_reg rx ry = true. breg_solve. *)

End test.